from __future__ import annotations

import math
from typing import Iterable, TYPE_CHECKING

import pygame

from ..entities_constants import (
    ZombieKind,
    ZOMBIE_SOLITARY_EVAL_INTERVAL_FRAMES,
    ZOMBIE_LINEFORMER_FOLLOW_DISTANCE,
    ZOMBIE_LINEFORMER_FOLLOW_TOLERANCE,
    ZOMBIE_LINEFORMER_SPEED_MULTIPLIER,
    ZOMBIE_SIGHT_RANGE,
    ZOMBIE_SPEED,
    ZOMBIE_TRACKER_FAR_SCENT_RADIUS,
    ZOMBIE_TRACKER_NEWER_FOOTPRINT_MS,
    ZOMBIE_TRACKER_RELOCK_DELAY_MS,
    ZOMBIE_TRACKER_SCENT_RADIUS,
    ZOMBIE_TRACKER_SCENT_TOP_K,
    ZOMBIE_TRACKER_SIGHT_RANGE,
    ZOMBIE_WALL_HUG_LOST_WALL_MS,
    ZOMBIE_WALL_HUG_PROBE_ANGLE_DEG,
    ZOMBIE_WALL_HUG_PROBE_ANGLE_PERP_DEG,
    ZOMBIE_WALL_HUG_TARGET_GAP,
    ZOMBIE_WALL_HUG_TURN_STEP_DEG,
)
from ..gameplay.constants import FOOTPRINT_STEP_DISTANCE
from ..rng import get_rng
from .movement import _circle_rect_collision

if TYPE_CHECKING:
    from ..models import Footprint, LevelLayout
    from . import Zombie

RNG = get_rng()


def _zombie_lineformer_train_head_movement(
    zombie: "Zombie",
    cell_size: int,
    layout: "LevelLayout",
    _player_center: tuple[float, float],
    nearby_zombies: Iterable["Zombie"],
    _footprints: list["Footprint"],
    *,
    now_ms: int,
) -> tuple[float, float]:
    target_pos = zombie.lineformer_target_pos
    if target_pos is None:
        return _zombie_wander_movement(
            zombie,
            cell_size,
            layout,
            now_ms=now_ms,
        )
    dx = target_pos[0] - zombie.x
    dy = target_pos[1] - zombie.y
    distance_sq = dx * dx + dy * dy
    target_id = zombie.lineformer_follow_target_id
    if target_id is not None:
        target_entity = next(
            (
                other
                for other in nearby_zombies
                if getattr(other, "lineformer_id", None) == target_id
            ),
            None,
        )
        if target_entity is not None:
            target_radius = float(
                getattr(
                    target_entity,
                    "collision_radius",
                    getattr(target_entity, "radius", zombie.collision_radius),
                )
            )
            min_distance = zombie.collision_radius + target_radius
            if distance_sq <= min_distance * min_distance:
                away_dx = zombie.x - float(getattr(target_entity, "x", target_pos[0]))
                away_dy = zombie.y - float(getattr(target_entity, "y", target_pos[1]))
                away_dist = math.hypot(away_dx, away_dy)
                if away_dist <= 1e-6:
                    angle = RNG.uniform(0.0, math.tau)
                    away_dx, away_dy = math.cos(angle), math.sin(angle)
                    away_dist = 1.0
                repel_speed = zombie.speed * ZOMBIE_LINEFORMER_SPEED_MULTIPLIER
                return (
                    (away_dx / away_dist) * repel_speed,
                    (away_dy / away_dist) * repel_speed,
                )
    follow_max = ZOMBIE_LINEFORMER_FOLLOW_DISTANCE + ZOMBIE_LINEFORMER_FOLLOW_TOLERANCE
    if distance_sq <= follow_max * follow_max:
        return 0.0, 0.0
    move_x, move_y = _zombie_move_toward(zombie, target_pos)
    return (
        move_x * ZOMBIE_LINEFORMER_SPEED_MULTIPLIER,
        move_y * ZOMBIE_LINEFORMER_SPEED_MULTIPLIER,
    )


def _zombie_tracker_movement(
    zombie: "Zombie",
    cell_size: int,
    layout: "LevelLayout",
    player_center: tuple[float, float],
    _nearby_zombies: Iterable["Zombie"],
    footprints: list["Footprint"],
    *,
    now_ms: int,
) -> tuple[float, float]:
    now = now_ms
    is_in_sight = zombie._update_mode(player_center, ZOMBIE_TRACKER_SIGHT_RANGE)
    if not is_in_sight:
        if zombie.tracker_force_wander:
            last_target_time = zombie.tracker_target_time
            if last_target_time is None:
                last_target_time = now
            zombie.tracker_relock_after_time = (
                last_target_time + ZOMBIE_TRACKER_RELOCK_DELAY_MS
            )
            zombie.tracker_target_pos = None
            return _zombie_wander_movement(
                zombie,
                cell_size,
                layout,
                now_ms=now,
            )
        _zombie_update_tracker_target(
            zombie,
            footprints,
            layout,
            cell_size=cell_size,
            now_ms=now,
        )
        if zombie.tracker_target_pos is not None:
            return _zombie_move_toward(zombie, zombie.tracker_target_pos)
        return _zombie_wander_movement(
            zombie,
            cell_size,
            layout,
            now_ms=now,
        )
    return _zombie_move_toward(zombie, player_center)


def _zombie_wall_hug_wall_distance(
    zombie: "Zombie",
    angle: float,
    max_distance: float,
    *,
    blocked_cells: set[tuple[int, int]],
    cell_size: int,
    grid_cols: int,
    grid_rows: int,
) -> float:
    """Approximate distance to nearest blocked cell along the ray."""
    if cell_size <= 0 or not blocked_cells:
        return max_distance

    direction_x = math.cos(angle)
    direction_y = math.sin(angle)
    radius = zombie.collision_radius

    def _hits_blocked(sample_dist: float) -> bool:
        sample_x = zombie.x + direction_x * sample_dist
        sample_y = zombie.y + direction_y * sample_dist
        min_cell_x = max(0, int((sample_x - radius) // cell_size))
        max_cell_x = min(grid_cols - 1, int((sample_x + radius) // cell_size))
        min_cell_y = max(0, int((sample_y - radius) // cell_size))
        max_cell_y = min(grid_rows - 1, int((sample_y + radius) // cell_size))
        for cy in range(min_cell_y, max_cell_y + 1):
            for cx in range(min_cell_x, max_cell_x + 1):
                if (cx, cy) not in blocked_cells:
                    continue
                rect = pygame.Rect(cx * cell_size, cy * cell_size, cell_size, cell_size)
                if _circle_rect_collision((sample_x, sample_y), radius, rect):
                    return True
        return False

    step = max(1.0, min(cell_size * 0.25, max_distance))
    prev_dist = 0.0
    scan_dist = 0.0
    while scan_dist <= max_distance:
        if _hits_blocked(scan_dist):
            lo, hi = prev_dist, scan_dist
            for _ in range(5):
                mid = (lo + hi) * 0.5
                if _hits_blocked(mid):
                    hi = mid
                else:
                    lo = mid
            return hi
        prev_dist = scan_dist
        scan_dist += step
    return max_distance


def _wall_hug_blocked_cells(layout: "LevelLayout") -> set[tuple[int, int]]:
    blocked = set(layout.wall_cells)
    blocked.update(layout.outer_wall_cells)
    blocked.update(layout.steel_beam_cells)
    blocked.update(layout.fire_floor_cells)
    return blocked


def _zombie_wall_hug_movement(
    zombie: "Zombie",
    cell_size: int,
    layout: "LevelLayout",
    player_center: tuple[float, float],
    _nearby_zombies: Iterable["Zombie"],
    _footprints: list["Footprint"],
    now_ms: int,
) -> tuple[float, float]:
    now = now_ms
    is_in_sight = zombie._update_mode(player_center, ZOMBIE_TRACKER_SIGHT_RANGE)
    if zombie.wall_hug_angle is None:
        zombie.wall_hug_angle = zombie.wander_angle

    # Speed-based scaling factors
    speed_ratio = zombie.speed / ZOMBIE_SPEED if ZOMBIE_SPEED > 0 else 1.0

    # Non-linear scaling for sensor distance based on cell size.
    dynamic_sensor_dist = cell_size * (0.4 + 0.005 * cell_size)
    dynamic_sensor_dist *= 0.8 + 0.2 * speed_ratio

    # Angles for Asymmetrical Probes
    # 1. Forward (0)
    # 2. Side Diagonal (default 55, scales with speed)
    # 3. Side Perpendicular (90)
    angle_side_deg = ZOMBIE_WALL_HUG_PROBE_ANGLE_DEG * (speed_ratio**-0.2)
    angle_perp_deg = ZOMBIE_WALL_HUG_PROBE_ANGLE_PERP_DEG

    probe_offset_side = math.radians(min(75.0, angle_side_deg))
    probe_offset_perp = math.radians(angle_perp_deg)

    sensor_distance = dynamic_sensor_dist + zombie.collision_radius
    target_gap_diagonal = ZOMBIE_WALL_HUG_TARGET_GAP / math.cos(probe_offset_side)
    blocked_cells = _wall_hug_blocked_cells(layout)

    if zombie.wall_hug_side == 0:
        # Initial side discovery (still symmetrical for the very first frame)
        left_angle = zombie.wall_hug_angle + probe_offset_side
        right_angle = zombie.wall_hug_angle - probe_offset_side
        left_dist = _zombie_wall_hug_wall_distance(
            zombie,
            left_angle,
            sensor_distance,
            blocked_cells=blocked_cells,
            cell_size=cell_size,
            grid_cols=layout.grid_cols,
            grid_rows=layout.grid_rows,
        )
        right_dist = _zombie_wall_hug_wall_distance(
            zombie,
            right_angle,
            sensor_distance,
            blocked_cells=blocked_cells,
            cell_size=cell_size,
            grid_cols=layout.grid_cols,
            grid_rows=layout.grid_rows,
        )
        forward_dist = _zombie_wall_hug_wall_distance(
            zombie,
            zombie.wall_hug_angle,
            sensor_distance,
            blocked_cells=blocked_cells,
            cell_size=cell_size,
            grid_cols=layout.grid_cols,
            grid_rows=layout.grid_rows,
        )

        if (
            left_dist < sensor_distance
            or right_dist < sensor_distance
            or forward_dist < sensor_distance
        ):
            if left_dist <= right_dist:
                zombie.wall_hug_side = 1.0
            else:
                zombie.wall_hug_side = -1.0
            zombie.wall_hug_last_wall_time = now
        else:
            if is_in_sight:
                return _zombie_move_toward(zombie, player_center)
            return _zombie_wander_movement(zombie, cell_size, layout, now_ms=now_ms)

    # Asymmetrical Probing while following
    side_angle = zombie.wall_hug_angle + zombie.wall_hug_side * probe_offset_side
    perp_angle = zombie.wall_hug_angle + zombie.wall_hug_side * probe_offset_perp

    side_dist = _zombie_wall_hug_wall_distance(
        zombie,
        side_angle,
        sensor_distance,
        blocked_cells=blocked_cells,
        cell_size=cell_size,
        grid_cols=layout.grid_cols,
        grid_rows=layout.grid_rows,
    )
    perp_dist = _zombie_wall_hug_wall_distance(
        zombie,
        perp_angle,
        sensor_distance,
        blocked_cells=blocked_cells,
        cell_size=cell_size,
        grid_cols=layout.grid_cols,
        grid_rows=layout.grid_rows,
    )
    forward_dist = _zombie_wall_hug_wall_distance(
        zombie,
        zombie.wall_hug_angle,
        sensor_distance,
        blocked_cells=blocked_cells,
        cell_size=cell_size,
        grid_cols=layout.grid_cols,
        grid_rows=layout.grid_rows,
    )

    side_has_wall = side_dist < sensor_distance
    perp_has_wall = perp_dist < sensor_distance
    forward_has_wall = forward_dist < sensor_distance

    wall_recent = (
        zombie.wall_hug_last_wall_time is not None
        and now - zombie.wall_hug_last_wall_time <= ZOMBIE_WALL_HUG_LOST_WALL_MS
    )
    if is_in_sight:
        return _zombie_move_toward(zombie, player_center)

    turn_step = math.radians(ZOMBIE_WALL_HUG_TURN_STEP_DEG * speed_ratio)

    if side_has_wall or perp_has_wall or forward_has_wall:
        zombie.wall_hug_last_wall_time = now

    # Steering Logic
    if perp_has_wall:
        # Obstacle is definitely there. Use diagonal for distance tracking.
        zombie.wall_hug_last_side_has_wall = True
        gap_error = target_gap_diagonal - side_dist
        if abs(gap_error) > 0.1:
            ratio = min(1.0, abs(gap_error) / target_gap_diagonal)
            turn = turn_step * ratio
            if gap_error > 0:
                zombie.wall_hug_angle -= zombie.wall_hug_side * turn
            else:
                zombie.wall_hug_angle += zombie.wall_hug_side * turn

        if forward_dist < ZOMBIE_WALL_HUG_TARGET_GAP:
            # Obstacle ahead! Turn sharply.
            zombie.wall_hug_angle -= zombie.wall_hug_side * (turn_step * 2.0)
    else:
        # PERPENDICULAR PROBE LOST WALL!
        # This means we either hit an outer corner or a GAP.
        zombie.wall_hug_last_side_has_wall = False
        if forward_has_wall:
            # Dead end? Turn away from forward wall.
            zombie.wall_hug_angle -= zombie.wall_hug_side * turn_step
        elif wall_recent:
            # Just lost it - turn sharply into the potential gap.
            zombie.wall_hug_angle += zombie.wall_hug_side * (turn_step * 1.2)
        else:
            # Completely lost - wander.
            zombie.wall_hug_angle += zombie.wall_hug_side * (math.pi / 2.0)
            zombie.wall_hug_side = 0.0
    zombie.wall_hug_angle %= math.tau

    move_x = math.cos(zombie.wall_hug_angle) * zombie.speed
    move_y = math.sin(zombie.wall_hug_angle) * zombie.speed
    return move_x, move_y


def _zombie_normal_movement(
    zombie: "Zombie",
    cell_size: int,
    layout: "LevelLayout",
    player_center: tuple[float, float],
    _nearby_zombies: Iterable["Zombie"],
    _footprints: list["Footprint"],
    now_ms: int,
) -> tuple[float, float]:
    is_in_sight = zombie._update_mode(player_center, ZOMBIE_SIGHT_RANGE)
    if not is_in_sight:
        return _zombie_wander_movement(
            zombie,
            cell_size,
            layout,
            now_ms=now_ms,
        )
    return _zombie_move_toward(zombie, player_center)


def _zombie_solitary_movement(
    zombie: "Zombie",
    cell_size: int,
    _layout: "LevelLayout",
    player_center: tuple[float, float],
    nearby_zombies: Iterable["Zombie"],
    _footprints: list["Footprint"],
    *,
    now_ms: int,
) -> tuple[float, float]:
    del now_ms
    if cell_size <= 0:
        return 0.0, 0.0
    zombie_weight = 3
    player_weight = 1
    if zombie.solitary_eval_frame_counter <= 0:
        zombie.solitary_eval_frame_counter = ZOMBIE_SOLITARY_EVAL_INTERVAL_FRAMES

        self_cell_x = int(zombie.x // cell_size)
        self_cell_y = int(zombie.y // cell_size)
        up_count = 0
        down_count = 0
        left_count = 0
        right_count = 0

        for other in nearby_zombies:
            if other is zombie or not other.alive():
                continue
            if bool(getattr(other, "is_trapped", False)):
                continue
            if getattr(other, "kind", None) == ZombieKind.DOG:
                continue
            other_cell_x = int(other.x // cell_size)
            other_cell_y = int(other.y // cell_size)
            dx = other_cell_x - self_cell_x
            dy = other_cell_y - self_cell_y
            if abs(dx) > 1 or abs(dy) > 1:
                continue
            if dy == -1 and abs(dx) <= 1:
                up_count += zombie_weight
            if dy == 1 and abs(dx) <= 1:
                down_count += zombie_weight
            if dx == -1 and abs(dy) <= 1:
                left_count += zombie_weight
            if dx == 1 and abs(dy) <= 1:
                right_count += zombie_weight
        player_cell_x = int(player_center[0] // cell_size)
        player_cell_y = int(player_center[1] // cell_size)
        player_dx = player_cell_x - self_cell_x
        player_dy = player_cell_y - self_cell_y
        if abs(player_dx) <= 1 and abs(player_dy) <= 1:
            if player_dy == -1 and abs(player_dx) <= 1:
                up_count += player_weight
            if player_dy == 1 and abs(player_dx) <= 1:
                down_count += player_weight
            if player_dx == -1 and abs(player_dy) <= 1:
                left_count += player_weight
            if player_dx == 1 and abs(player_dy) <= 1:
                right_count += player_weight

        dir_x = 0
        dir_y = 0
        if up_count < down_count:
            dir_y = -1
        elif down_count < up_count:
            dir_y = 1
        if left_count < right_count:
            dir_x = -1
        elif right_count < left_count:
            dir_x = 1

        proposed_move: tuple[int, int] | None
        if dir_x == 0 and dir_y == 0:
            proposed_move = None
        else:
            proposed_move = (dir_x, dir_y)
        prev_move = zombie.solitary_previous_move
        if (
            proposed_move is not None
            and prev_move is not None
            and proposed_move == (-prev_move[0], -prev_move[1])
        ):
            proposed_move = None
        zombie.solitary_committed_move = proposed_move
        if proposed_move is not None:
            zombie.solitary_previous_move = proposed_move

    zombie.solitary_eval_frame_counter -= 1
    if zombie.solitary_committed_move is None:
        return 0.0, 0.0
    step_speed = zombie.speed * 2.2
    move_x = float(zombie.solitary_committed_move[0]) * step_speed
    move_y = float(zombie.solitary_committed_move[1]) * step_speed
    if move_x != 0.0 and move_y != 0.0:
        norm = math.sqrt(2.0)
        move_x /= norm
        move_y /= norm
    return move_x, move_y


def _zombie_update_tracker_target(
    zombie: "Zombie",
    footprints: list["Footprint"],
    layout: "LevelLayout",
    *,
    cell_size: int,
    now_ms: int,
) -> None:
    # footprints are ordered oldest -> newest by time.
    now = now_ms
    if now - zombie.tracker_last_scan_time < zombie.tracker_scan_interval_ms:
        return
    zombie.tracker_last_scan_time = now
    if not footprints:
        zombie.tracker_target_pos = None
        return
    last_target_time = zombie.tracker_target_time
    far_radius_sq = ZOMBIE_TRACKER_FAR_SCENT_RADIUS * ZOMBIE_TRACKER_FAR_SCENT_RADIUS
    relock_after = zombie.tracker_relock_after_time
    far_candidates: list[tuple[float, Footprint]] = []
    for fp in footprints:
        dx = fp.pos[0] - zombie.x
        dy = fp.pos[1] - zombie.y
        d2 = dx * dx + dy * dy
        if d2 <= far_radius_sq:
            far_candidates.append((d2, fp))
    if not far_candidates:
        return
    latest_fp_time = far_candidates[-1][1].time
    use_far_scan = last_target_time is None or (
        latest_fp_time is not None
        and latest_fp_time - last_target_time >= ZOMBIE_TRACKER_NEWER_FOOTPRINT_MS
    )
    scan_radius = (
        ZOMBIE_TRACKER_FAR_SCENT_RADIUS if use_far_scan else ZOMBIE_TRACKER_SCENT_RADIUS
    )
    scent_radius_sq = scan_radius * scan_radius
    min_target_dist_sq = (FOOTPRINT_STEP_DISTANCE * 0.5) ** 2

    newer: list[Footprint] = []
    for d2, fp in far_candidates:
        pos = fp.pos
        fp_time = fp.time
        if relock_after is not None and fp_time < relock_after:
            continue
        if d2 <= min_target_dist_sq:
            continue
        if d2 <= scent_radius_sq:
            if last_target_time is None or fp_time > last_target_time:
                newer.append(fp)

    if not newer:
        return

    newer.sort(key=lambda fp: fp.time)

    if use_far_scan or last_target_time is None:
        candidates = list(reversed(newer))[:ZOMBIE_TRACKER_SCENT_TOP_K]
    else:
        newer_threshold = last_target_time + ZOMBIE_TRACKER_NEWER_FOOTPRINT_MS
        very_new = [fp for fp in newer if fp.time >= newer_threshold]
        if very_new:
            candidates = list(reversed(very_new))[:ZOMBIE_TRACKER_SCENT_TOP_K]
        else:
            candidates = newer[:ZOMBIE_TRACKER_SCENT_TOP_K]

    blocked_cells = _tracker_blocked_cells(layout)
    for fp in candidates:
        pos = fp.pos
        fp_time = fp.time
        if _line_of_sight_clear_cells(
            (zombie.x, zombie.y),
            pos,
            blocked_cells=blocked_cells,
            cell_size=cell_size,
            grid_cols=layout.grid_cols,
            grid_rows=layout.grid_rows,
        ):
            zombie.tracker_target_pos = pos
            zombie.tracker_target_time = fp_time
            if relock_after is not None and fp_time >= relock_after:
                zombie.tracker_relock_after_time = None
            return

    if (
        zombie.tracker_target_pos is not None
        and (zombie.x - zombie.tracker_target_pos[0]) ** 2
        + (zombie.y - zombie.tracker_target_pos[1]) ** 2
        > min_target_dist_sq
    ):
        return

    if last_target_time is None:
        return

    next_fp = newer[0]
    zombie.tracker_target_pos = next_fp.pos
    zombie.tracker_target_time = next_fp.time
    if relock_after is not None and next_fp.time >= relock_after:
        zombie.tracker_relock_after_time = None
    return


def _line_of_sight_clear_cells(
    start: tuple[float, float],
    end: tuple[float, float],
    *,
    blocked_cells: set[tuple[int, int]],
    cell_size: int,
    grid_cols: int,
    grid_rows: int,
) -> bool:
    if cell_size <= 0 or not blocked_cells:
        return True
    dx = end[0] - start[0]
    dy = end[1] - start[1]
    dist = math.hypot(dx, dy)
    if dist <= 1e-6:
        return True
    step = max(1.0, min(cell_size * 0.5, dist))
    samples = max(1, int(math.ceil(dist / step)))
    for i in range(samples + 1):
        t = i / samples
        x = start[0] + dx * t
        y = start[1] + dy * t
        cx = int(x // cell_size)
        cy = int(y // cell_size)
        if cx < 0 or cy < 0 or cx >= grid_cols or cy >= grid_rows:
            continue
        if (cx, cy) in blocked_cells:
            return False
    return True


def _tracker_blocked_cells(layout: "LevelLayout") -> set[tuple[int, int]]:
    blocked = set(layout.wall_cells)
    blocked.update(layout.outer_wall_cells)
    blocked.update(layout.steel_beam_cells)
    return blocked


def _zombie_wander_movement(
    zombie: "Zombie",
    cell_size: int,
    layout: "LevelLayout",
    *,
    now_ms: int,
) -> None:
    grid_cols = layout.grid_cols
    grid_rows = layout.grid_rows
    outer_wall_cells = layout.outer_wall_cells
    pitfall_cells = layout.pitfall_cells
    fire_floor_cells = layout.fire_floor_cells
    now = now_ms
    changed_angle = False
    if now - zombie.last_wander_change_time > zombie.wander_change_interval:
        zombie.wander_angle = RNG.uniform(0, math.tau)
        zombie.last_wander_change_time = now
        jitter = RNG.randint(-500, 500)
        zombie.wander_change_interval = max(0, zombie.wander_interval_ms + jitter)
        changed_angle = True

    cell_x = int(zombie.x // cell_size)
    cell_y = int(zombie.y // cell_size)
    at_x_edge = cell_x in (0, grid_cols - 1)
    at_y_edge = cell_y in (0, grid_rows - 1)
    if changed_angle and (at_x_edge or at_y_edge):
        cos_angle = math.cos(zombie.wander_angle)
        sin_angle = math.sin(zombie.wander_angle)
        outward = (
            (cell_x == 0 and cos_angle < 0)
            or (cell_x == grid_cols - 1 and cos_angle > 0)
            or (cell_y == 0 and sin_angle < 0)
            or (cell_y == grid_rows - 1 and sin_angle > 0)
        )
        if outward:
            if RNG.random() < 0.5:
                zombie.wander_angle = (zombie.wander_angle + math.pi) % math.tau

    if at_x_edge or at_y_edge:
        if at_x_edge:
            inward_cell = (1, cell_y) if cell_x == 0 else (grid_cols - 2, cell_y)
            if inward_cell not in outer_wall_cells:
                target_x = (inward_cell[0] + 0.5) * cell_size
                target_y = (inward_cell[1] + 0.5) * cell_size
                return _zombie_move_toward(zombie, (target_x, target_y))
        if at_y_edge:
            inward_cell = (cell_x, 1) if cell_y == 0 else (cell_x, grid_rows - 2)
            if inward_cell not in outer_wall_cells:
                target_x = (inward_cell[0] + 0.5) * cell_size
                target_y = (inward_cell[1] + 0.5) * cell_size
                return _zombie_move_toward(zombie, (target_x, target_y))

    def _base_move() -> tuple[float, float]:
        return (
            math.cos(zombie.wander_angle) * zombie.speed,
            math.sin(zombie.wander_angle) * zombie.speed,
        )

    move_x, move_y = _base_move()
    if pitfall_cells is not None:
        avoid_x, avoid_y = zombie._avoid_pitfalls(pitfall_cells, cell_size)
        move_x += avoid_x
        move_y += avoid_y
        if cell_size > 0:
            next_x = zombie.x + move_x
            next_y = zombie.y + move_y
            next_cell = (int(next_x // cell_size), int(next_y // cell_size))
            if next_cell in pitfall_cells:
                zombie.wander_angle = (zombie.wander_angle + math.pi) % math.tau
                move_x, move_y = _base_move()
                avoid_x, avoid_y = zombie._avoid_pitfalls(pitfall_cells, cell_size)
                move_x += avoid_x
                move_y += avoid_y
                next_x = zombie.x + move_x
                next_y = zombie.y + move_y
                next_cell = (int(next_x // cell_size), int(next_y // cell_size))
                if next_cell in pitfall_cells:
                    return 0.0, 0.0
    if fire_floor_cells and cell_size > 0:
        next_x = zombie.x + move_x
        next_y = zombie.y + move_y
        next_cell = (int(next_x // cell_size), int(next_y // cell_size))
        if next_cell in fire_floor_cells:
            zombie.wander_angle = (zombie.wander_angle + math.pi) % math.tau
            move_x, move_y = _base_move()
            if pitfall_cells is not None:
                avoid_x, avoid_y = zombie._avoid_pitfalls(pitfall_cells, cell_size)
                move_x += avoid_x
                move_y += avoid_y
            next_x = zombie.x + move_x
            next_y = zombie.y + move_y
            next_cell = (int(next_x // cell_size), int(next_y // cell_size))
            if next_cell in fire_floor_cells:
                return 0.0, 0.0
    return move_x, move_y


def _zombie_move_toward(
    zombie: "Zombie", target: tuple[float, float]
) -> tuple[float, float]:
    dx = target[0] - zombie.x
    dy = target[1] - zombie.y
    dist = math.hypot(dx, dy)
    if dist <= 0:
        return 0.0, 0.0
    return (dx / dist) * zombie.speed, (dy / dist) * zombie.speed



from __future__ import annotations

from typing import TYPE_CHECKING

import pygame
from pygame import sprite, surface

from .colors import (
    BLACK,
    BLUE,
    FOOTPRINT_COLOR,
    LIGHT_GRAY,
    RED,
    YELLOW,
    WHITE,
    get_environment_palette,
)
from .entities import (
    Car,
    EmptyFuelCan,
    Flashlight,
    FuelCan,
    FuelStation,
    Player,
    PatrolBot,
    Shoes,
    SpikyPlant,
    SteelBeam,
    Survivor,
    TrappedZombie,
    Wall,
    Zombie,
    ZombieDog,
)
from .models import Footprint, GameData
from .render_assets import RenderAssets, resolve_steel_beam_colors, resolve_wall_colors
from .render_constants import (
    FIRE_FLOOR_FLAME_COLORS,
    LINEFORMER_MARKER_OVERVIEW_COLOR,
    MOVING_FLOOR_OVERVIEW_COLOR,
    TRAPPED_ZOMBIE_OVERVIEW_COLOR,
)
from .entities_constants import PATROL_BOT_COLLISION_RADIUS
from .render.hud import _draw_status_bar, _get_fog_scale
from .render.puddle import draw_puddle_rings, get_puddle_wave_color

if TYPE_CHECKING:  # pragma: no cover - typing-only imports
    from .gameplay.lineformer_trains import LineformerTrainManager


def compute_floor_cells(
    *,
    cols: int,
    rows: int,
    wall_cells: set[tuple[int, int]],
    outer_wall_cells: set[tuple[int, int]],
    pitfall_cells: set[tuple[int, int]],
) -> set[tuple[int, int]]:
    """Return floor cells for the minimap base pass."""
    # The layout wall sets are updated when walls are destroyed, so removing
    # those cells here makes the minimap treat destroyed walls as floor.
    blocked = wall_cells | outer_wall_cells | pitfall_cells
    return {(x, y) for y in range(rows) for x in range(cols) if (x, y) not in blocked}


def _draw_spiky_plant_spike_mark(
    surface: surface.Surface,
    center: tuple[int, int],
    radius: int,
) -> None:
    body_fill_color = (64, 132, 64)
    body_outline_color = (90, 180, 90)
    spike_color = (90, 180, 90)
    body_radius = max(2, radius - 2)
    pygame.draw.circle(surface, body_fill_color, center, body_radius)
    pygame.draw.circle(surface, body_outline_color, center, body_radius, width=1)
    inner = max(1, radius - 2)
    outer = max(inner + 1, radius + 1)
    for i in range(8):
        angle = i * (360 / 8)
        direction = pygame.Vector2(1, 0).rotate(angle)
        start = (
            int(center[0] + direction.x * inner),
            int(center[1] + direction.y * inner),
        )
        end = (
            int(center[0] + direction.x * outer),
            int(center[1] + direction.y * outer),
        )
        pygame.draw.line(surface, spike_color, start, end, width=2)


def _draw_overview_floor_layers(
    surface: surface.Surface,
    *,
    cell_size: int,
    floor_cells: set[tuple[int, int]],
    floor_color: tuple[int, int, int],
    fall_floor: tuple[int, int, int],
    fall_spawn_cells: set[tuple[int, int]] | None,
    moving_floor_cells: dict[tuple[int, int], object] | None,
    fire_floor_cells: set[tuple[int, int]] | None,
    puddle_cells: set[tuple[int, int]] | None,
    zombie_contaminated_cells: set[tuple[int, int]] | None,
) -> None:
    if cell_size <= 0:
        return

    def _cell_rect(x: int, y: int) -> pygame.Rect:
        return pygame.Rect(x * cell_size, y * cell_size, cell_size, cell_size)

    for x, y in floor_cells:
        pygame.draw.rect(surface, floor_color, _cell_rect(x, y))
    if fall_spawn_cells:
        for x, y in fall_spawn_cells:
            pygame.draw.rect(surface, fall_floor, _cell_rect(x, y))
    if moving_floor_cells:
        for x, y in moving_floor_cells.keys():
            pygame.draw.rect(surface, MOVING_FLOOR_OVERVIEW_COLOR, _cell_rect(x, y))
    if fire_floor_cells:
        fire_diamond = tuple(
            max(0, int(channel * 0.35)) for channel in FIRE_FLOOR_FLAME_COLORS[0]
        )
        for x, y in fire_floor_cells:
            cell_rect = _cell_rect(x, y)
            inset = max(1, int(cell_size * 0.22))
            cx = cell_rect.centerx
            cy = cell_rect.centery
            diamond = [
                (cx, cell_rect.top + inset),
                (cell_rect.right - inset, cy),
                (cx, cell_rect.bottom - inset),
                (cell_rect.left + inset, cy),
            ]
            pygame.draw.polygon(surface, fire_diamond, diamond)
    if puddle_cells:
        puddle_wave_color = get_puddle_wave_color(alpha=None)
        for x, y in puddle_cells:
            draw_puddle_rings(
                surface,
                rect=_cell_rect(x, y),
                phase=0,
                color=puddle_wave_color,
                width=1,
            )
    if zombie_contaminated_cells:
        for x, y in zombie_contaminated_cells:
            inset = max(1, int(cell_size * 0.05))
            cell_rect = _cell_rect(x, y)
            pygame.draw.rect(
                surface,
                RED,
                pygame.Rect(
                    cell_rect.x + inset,
                    cell_rect.y + inset,
                    max(1, cell_size - inset * 2),
                    max(1, cell_size - inset * 2),
                ),
                width=1,
            )


def _draw_overview_walls(
    surface: surface.Surface,
    *,
    wall_group: sprite.Group,
    floor_color: tuple[int, int, int],
    palette: object,
) -> None:
    for wall in wall_group:
        if wall.max_health > 0:
            health_ratio = max(0.0, min(1.0, wall.health / wall.max_health))
        else:
            health_ratio = 0.0
        if isinstance(wall, Wall):
            if health_ratio <= 0.0:
                pygame.draw.rect(surface, floor_color, wall.rect)
            else:
                fill_color, _ = resolve_wall_colors(
                    health_ratio=health_ratio,
                    palette_category=wall.palette_category,
                    palette=palette,
                )
                pygame.draw.rect(surface, fill_color, wall.rect)
        elif isinstance(wall, SteelBeam):
            if health_ratio <= 0.0:
                pygame.draw.rect(surface, floor_color, wall.rect)
            else:
                fill_color, _ = resolve_steel_beam_colors(
                    health_ratio=health_ratio,
                    palette=palette,
                )
                pygame.draw.rect(surface, fill_color, wall.rect)


def _draw_overview_footprints(
    surface: surface.Surface,
    *,
    assets: RenderAssets,
    footprints: list[Footprint],
    now_ms: int,
) -> None:
    for fp in footprints:
        if not fp.visible:
            continue
        age = now_ms - fp.time
        fade = 1 - (age / assets.footprint_lifetime_ms)
        fade = max(assets.footprint_min_fade, fade)
        color = tuple(max(0, min(255, int(c * fade))) for c in FOOTPRINT_COLOR)
        pygame.draw.circle(
            surface,
            color,
            (int(fp.pos[0]), int(fp.pos[1])),
            assets.footprint_overview_radius,
        )


def _draw_overview_items(
    surface: surface.Surface,
    *,
    fuel: FuelCan | None,
    empty_fuel_can: EmptyFuelCan | None,
    fuel_station: FuelStation | None,
    flashlights: list[Flashlight] | None,
    shoes: list[Shoes] | None,
) -> None:
    if fuel and fuel.alive():
        pygame.draw.rect(surface, YELLOW, fuel.rect, border_radius=3)
        pygame.draw.rect(surface, BLACK, fuel.rect, width=2, border_radius=3)
    if empty_fuel_can and empty_fuel_can.alive():
        pygame.draw.rect(surface, LIGHT_GRAY, empty_fuel_can.rect, border_radius=3)
        pygame.draw.rect(surface, BLACK, empty_fuel_can.rect, width=2, border_radius=3)
    if fuel_station and fuel_station.alive():
        pygame.draw.rect(surface, YELLOW, fuel_station.rect, border_radius=2)
        pygame.draw.rect(surface, BLACK, fuel_station.rect, width=2, border_radius=2)
    if flashlights:
        for flashlight in flashlights:
            if flashlight.alive():
                pygame.draw.rect(surface, YELLOW, flashlight.rect, border_radius=2)
                pygame.draw.rect(
                    surface, BLACK, flashlight.rect, width=2, border_radius=2
                )
    if shoes:
        for item in shoes:
            if item.alive():
                surface.blit(item.image, item.rect)


def _draw_overview_humanoids(
    surface: surface.Surface,
    *,
    assets: RenderAssets,
    player: Player | None,
    survivors: list[Survivor] | None,
    buddies: list[Survivor] | None,
    car: Car | None,
    waiting_cars: list[Car] | None,
    patrol_bots: list[PatrolBot] | None,
    spiky_plants: list[SpikyPlant] | None,
) -> None:
    if survivors:
        for survivor in survivors:
            if survivor.alive():
                pygame.draw.circle(
                    surface,
                    (220, 220, 255),
                    survivor.rect.center,
                    assets.player_radius * 2,
                )
    if player:
        pygame.draw.circle(surface, BLUE, player.rect.center, assets.player_radius * 2)
    if buddies:
        buddy_color = (0, 200, 70)
        for buddy in buddies:
            if buddy.alive() and not buddy.rescued:
                pygame.draw.circle(
                    surface, buddy_color, buddy.rect.center, assets.player_radius * 2
                )
    drawn_cars: list[Car] = []
    if car and car.alive():
        car_rect = car.image.get_rect(center=car.rect.center)
        surface.blit(car.image, car_rect)
        drawn_cars.append(car)
    if waiting_cars:
        for parked in waiting_cars:
            if not parked.alive() or parked in drawn_cars:
                continue
            parked_rect = parked.image.get_rect(center=parked.rect.center)
            surface.blit(parked.image, parked_rect)
    if patrol_bots:
        for bot in patrol_bots:
            if bot.alive():
                pygame.draw.circle(
                    surface,
                    (90, 45, 120),
                    bot.rect.center,
                    int(PATROL_BOT_COLLISION_RADIUS),
                )
    if spiky_plants:
        for hp in spiky_plants:
            if hp.alive():
                _draw_spiky_plant_spike_mark(
                    surface,
                    hp.rect.center,
                    max(2, int(hp.radius)),
                )


def _draw_overview_zombies(
    surface: surface.Surface,
    *,
    assets: RenderAssets,
    zombies: list[pygame.sprite.Sprite] | None,
    lineformer_trains: "LineformerTrainManager | None",
) -> None:
    if not zombies:
        return
    zombie_color = (200, 80, 80)
    zombie_radius = max(2, int(assets.player_radius * 1.2))
    for zombie in zombies:
        if not zombie.alive():
            continue
        if isinstance(zombie, (Zombie, ZombieDog)):
            pygame.draw.circle(
                surface,
                zombie_color,
                zombie.rect.center,
                zombie_radius,
            )
        elif isinstance(zombie, TrappedZombie):
            pygame.draw.circle(
                surface,
                TRAPPED_ZOMBIE_OVERVIEW_COLOR,
                zombie.rect.center,
                zombie_radius,
            )
    if lineformer_trains is not None:
        marker_radius = max(1, zombie_radius - 1)
        marker_draw_data = lineformer_trains.iter_marker_draw_data(zombies)
        for marker_x, marker_y, _ in marker_draw_data:
            pygame.draw.circle(
                surface,
                LINEFORMER_MARKER_OVERVIEW_COLOR,
                (int(marker_x), int(marker_y)),
                marker_radius,
            )


def draw_level_overview(
    assets: RenderAssets,
    surface: surface.Surface,
    wall_group: sprite.Group,
    floor_cells: set[tuple[int, int]],
    player: Player | None,
    car: Car | None,
    waiting_cars: list[Car] | None,
    footprints: list[Footprint],
    *,
    now_ms: int,
    fuel: FuelCan | None = None,
    empty_fuel_can: EmptyFuelCan | None = None,
    fuel_station: FuelStation | None = None,
    flashlights: list[Flashlight] | None = None,
    shoes: list[Shoes] | None = None,
    buddies: list[Survivor] | None = None,
    survivors: list[Survivor] | None = None,
    patrol_bots: list[PatrolBot] | None = None,
    spiky_plants: list[SpikyPlant] | None = None,
    zombies: list[pygame.sprite.Sprite] | None = None,
    lineformer_trains: "LineformerTrainManager | None" = None,
    fall_spawn_cells: set[tuple[int, int]] | None = None,
    moving_floor_cells: dict[tuple[int, int], object] | None = None,
    fire_floor_cells: set[tuple[int, int]] | None = None,
    puddle_cells: set[tuple[int, int]] | None = None,
    zombie_contaminated_cells: set[tuple[int, int]] | None = None,
    palette_key: str | None = None,
) -> None:
    palette = get_environment_palette(palette_key)
    base_floor = palette.floor_primary
    dark_floor = tuple(max(0, int(channel * 0.35)) for channel in base_floor)
    floor_color = tuple(max(0, int(channel * 0.65)) for channel in base_floor)
    fall_floor = tuple(
        max(0, int(channel * 0.55)) for channel in palette.fall_zone_primary
    )
    surface.fill(dark_floor)
    _draw_overview_floor_layers(
        surface,
        cell_size=assets.internal_wall_grid_snap,
        floor_cells=floor_cells,
        floor_color=floor_color,
        fall_floor=fall_floor,
        fall_spawn_cells=fall_spawn_cells,
        moving_floor_cells=moving_floor_cells,
        fire_floor_cells=fire_floor_cells,
        puddle_cells=puddle_cells,
        zombie_contaminated_cells=zombie_contaminated_cells,
    )
    _draw_overview_walls(
        surface,
        wall_group=wall_group,
        floor_color=floor_color,
        palette=palette,
    )
    _draw_overview_footprints(
        surface,
        assets=assets,
        footprints=footprints,
        now_ms=now_ms,
    )
    _draw_overview_items(
        surface,
        fuel=fuel,
        empty_fuel_can=empty_fuel_can,
        fuel_station=fuel_station,
        flashlights=flashlights,
        shoes=shoes,
    )
    _draw_overview_humanoids(
        surface,
        assets=assets,
        player=player,
        survivors=survivors,
        buddies=buddies,
        car=car,
        waiting_cars=waiting_cars,
        patrol_bots=patrol_bots,
        spiky_plants=spiky_plants,
    )
    _draw_overview_zombies(
        surface,
        assets=assets,
        zombies=zombies,
        lineformer_trains=lineformer_trains,
    )


def draw_debug_overview(
    assets: RenderAssets,
    screen: surface.Surface,
    overview_surface: surface.Surface,
    game_data: GameData,
    config: dict[str, object],
    *,
    screen_width: int,
    screen_height: int,
    show_debug_counts: bool = True,
) -> None:
    cell_size = assets.internal_wall_grid_snap
    floor_cells: set[tuple[int, int]] = set()
    if cell_size > 0:
        floor_cells = compute_floor_cells(
            cols=game_data.layout.grid_cols,
            rows=game_data.layout.grid_rows,
            wall_cells=game_data.layout.wall_cells,
            outer_wall_cells=game_data.layout.outer_wall_cells,
            pitfall_cells=game_data.layout.pitfall_cells,
        )
    footprints_enabled = bool(config.get("footprints", {}).get("enabled", True))
    footprints_to_draw = game_data.state.footprints if footprints_enabled else []
    draw_level_overview(
        assets,
        overview_surface,
        game_data.groups.wall_group,
        floor_cells,
        game_data.player,
        game_data.car,
        game_data.waiting_cars,
        footprints_to_draw,
        now_ms=game_data.state.clock.elapsed_ms,
        fuel=game_data.fuel,
        empty_fuel_can=game_data.empty_fuel_can,
        fuel_station=game_data.fuel_station,
        flashlights=game_data.flashlights or [],
        shoes=game_data.shoes or [],
        buddies=[
            survivor
            for survivor in game_data.groups.survivor_group
            if survivor.alive() and survivor.is_buddy and not survivor.rescued
        ],
        survivors=list(game_data.groups.survivor_group),
        patrol_bots=list(game_data.groups.patrol_bot_group),
        spiky_plants=list(game_data.spiky_plants.values()),
        zombies=list(game_data.groups.zombie_group),
        lineformer_trains=game_data.lineformer_trains,
        fall_spawn_cells=game_data.layout.fall_spawn_cells,
        moving_floor_cells=game_data.layout.moving_floor_cells,
        fire_floor_cells=game_data.layout.fire_floor_cells,
        puddle_cells=game_data.layout.puddle_cells,
        zombie_contaminated_cells=game_data.layout.zombie_contaminated_cells,
        palette_key=game_data.state.ambient_palette_key,
    )
    fov_target = None
    if (
        game_data.player
        and game_data.player.in_car
        and game_data.car
        and game_data.car.alive()
    ):
        fov_target = game_data.car
    elif game_data.player:
        fov_target = game_data.player
    if fov_target:
        fov_scale = _get_fog_scale(assets, game_data.state.flashlight_count)
        fov_radius = max(1, int(assets.fov_radius * fov_scale))
        pygame.draw.circle(
            overview_surface,
            (255, 255, 120),
            fov_target.rect.center,
            fov_radius,
            width=2,
        )
    cam_offset = game_data.camera.camera
    camera_rect = pygame.Rect(
        -cam_offset.x,
        -cam_offset.y,
        screen_width,
        screen_height,
    )
    pygame.draw.rect(overview_surface, WHITE, camera_rect, width=1)
    level_rect = game_data.layout.field_rect
    level_aspect = level_rect.width / max(1, level_rect.height)
    screen_aspect = screen_width / max(1, screen_height)
    if level_aspect > screen_aspect:
        scaled_w = screen_width - 40
        scaled_h = int(scaled_w / level_aspect)
    else:
        scaled_h = screen_height - 40
        scaled_w = int(scaled_h * level_aspect)
    scaled_w = max(1, scaled_w)
    scaled_h = max(1, scaled_h)
    scaled_overview = pygame.transform.smoothscale(
        overview_surface, (scaled_w, scaled_h)
    )
    screen.fill(BLACK)
    scaled_rect = scaled_overview.get_rect(
        center=(screen_width // 2, screen_height // 2)
    )
    screen.blit(
        scaled_overview,
        scaled_rect,
    )
    _draw_status_bar(
        screen,
        assets,
        config,
        stage=game_data.stage,
        seed=game_data.state.seed,
        debug_mode=show_debug_counts,
        zombie_group=game_data.groups.zombie_group,
        lineformer_marker_count=game_data.lineformer_trains.total_marker_count(),
        falling_spawn_carry=game_data.state.falling_spawn_carry,
        show_fps=game_data.state.show_fps,
    )



from __future__ import annotations

import random
from typing import Any

import pygame

from ..colors import get_environment_palette
from ..entities import ReinforcedWall, RubbleWall, SteelBeam, Wall
from ..entities_constants import (
    INTERNAL_WALL_BEVEL_DEPTH,
    INTERNAL_WALL_HEALTH,
    MovingFloorDirection,
    STEEL_BEAM_HEALTH,
)
from ..level_constants import DEFAULT_STEEL_BEAM_CHANCE
from ..render_assets import RUBBLE_ROTATION_DEG
from .constants import LAYER_WALLS, OUTER_WALL_HEALTH
from ..level_blueprints import (
    Blueprint,
    MapGenerationError,
    generate_random_blueprint,
    validate_connectivity,
)
from ..models import LevelLayout, Stage
from ..models import FuelMode
from ..rng import get_rng, seed_rng

__all__ = ["generate_level_from_blueprint", "MapGenerationError"]

RNG = get_rng()


def _rect_for_cell(x_idx: int, y_idx: int, cell_size: int) -> pygame.Rect:
    return pygame.Rect(
        x_idx * cell_size,
        y_idx * cell_size,
        cell_size,
        cell_size,
    )


def _expand_zone_cells(
    zones: list[tuple[int, int, int, int]],
    *,
    grid_cols: int,
    grid_rows: int,
) -> set[tuple[int, int]]:
    cells: set[tuple[int, int]] = set()
    for col, row, width, height in zones:
        if width <= 0 or height <= 0:
            continue
        start_x = max(0, col)
        start_y = max(0, row)
        end_x = min(grid_cols, col + width)
        end_y = min(grid_rows, row + height)
        for y in range(start_y, end_y):
            for x in range(start_x, end_x):
                cells.add((x, y))
    return cells


def _expand_moving_floor_cells(
    stage: Stage,
) -> dict[tuple[int, int], MovingFloorDirection]:
    directions: dict[str, MovingFloorDirection] = {
        "u": MovingFloorDirection.UP,
        "up": MovingFloorDirection.UP,
        "d": MovingFloorDirection.DOWN,
        "down": MovingFloorDirection.DOWN,
        "l": MovingFloorDirection.LEFT,
        "left": MovingFloorDirection.LEFT,
        "r": MovingFloorDirection.RIGHT,
        "right": MovingFloorDirection.RIGHT,
        "U": MovingFloorDirection.UP,
        "D": MovingFloorDirection.DOWN,
        "L": MovingFloorDirection.LEFT,
        "R": MovingFloorDirection.RIGHT,
    }
    moving_floor_cells: dict[tuple[int, int], MovingFloorDirection] = {}
    if stage.moving_floor_zones:
        for key, zones in stage.moving_floor_zones.items():
            direction = directions.get(str(key).lower(), directions.get(str(key)))
            if not direction or not zones:
                continue
            for cell in _expand_zone_cells(
                zones,
                grid_cols=stage.grid_cols,
                grid_rows=stage.grid_rows,
            ):
                moving_floor_cells[cell] = direction
    if stage.moving_floor_cells:
        for cell, direction in stage.moving_floor_cells.items():
            try:
                dir_enum = (
                    direction
                    if isinstance(direction, MovingFloorDirection)
                    else MovingFloorDirection(direction)
                )
            except ValueError:
                continue
            moving_floor_cells[cell] = dir_enum
    return moving_floor_cells


def generate_level_from_blueprint(
    stage: Stage,
    config: dict[str, Any],
    *,
    seed: int | None,
    ambient_palette_key: str | None,
) -> tuple[
    LevelLayout,
    dict[str, list[tuple[int, int]]],
    pygame.sprite.Group,
    pygame.sprite.LayeredUpdates,
    Blueprint,
]:
    """Build walls/spawn candidates/outside area from a blueprint grid."""
    wall_group = pygame.sprite.Group()
    all_sprites = pygame.sprite.LayeredUpdates()

    steel_conf = config.get("steel_beams", {})
    steel_enabled = steel_conf.get("enabled", False)

    base_moving_floor_cells = _expand_moving_floor_cells(stage)
    fuel_count = 0
    empty_fuel_can_count = 0
    fuel_station_count = 0
    if stage.fuel_mode < FuelMode.START_FULL and not stage.endurance_stage:
        if stage.fuel_mode == FuelMode.REFUEL_CHAIN:
            empty_fuel_can_count = max(1, int(stage.empty_fuel_can_spawn_count))
            fuel_station_count = max(1, int(stage.fuel_station_spawn_count))
        else:
            fuel_count = max(0, int(stage.fuel_spawn_count))
    flashlight_count = max(0, int(stage.initial_flashlight_count))
    shoes_count = max(0, int(stage.initial_shoes_count))

    steel_conf = config.get("steel_beams", {})
    try:
        steel_chance = float(steel_conf.get("chance", DEFAULT_STEEL_BEAM_CHANCE))
    except (TypeError, ValueError):
        steel_chance = DEFAULT_STEEL_BEAM_CHANCE

    # Connectivity validation and retries live here to keep generation shallow.
    blueprint_data = None
    for attempt in range(20):
        if seed is not None:
            seed_rng(seed + attempt)
        try:
            blueprint = generate_random_blueprint(
                steel_chance=steel_chance,
                cols=stage.grid_cols,
                rows=stage.grid_rows,
                exit_sides=stage.exit_sides,
                wall_algo=stage.wall_algorithm,
                pitfall_density=stage.pitfall_density,
                pitfall_zones=stage.pitfall_zones,
                fire_floor_density=stage.fire_floor_density,
                fire_floor_zones=stage.fire_floor_zones,
                metal_floor_density=stage.metal_floor_density,
                metal_floor_zones=stage.metal_floor_zones,
                reinforced_wall_density=stage.reinforced_wall_density,
                reinforced_wall_zones=stage.reinforced_wall_zones,
                moving_floor_cells=base_moving_floor_cells,
                fuel_count=fuel_count,
                empty_fuel_can_count=empty_fuel_can_count,
                fuel_station_count=fuel_station_count,
                flashlight_count=flashlight_count,
                shoes_count=shoes_count,
                spiky_plant_density=stage.spiky_plant_density,
                spiky_plant_zones=stage.spiky_plant_zones,
                puddle_density=stage.puddle_density,
                puddle_zones=stage.puddle_zones,
            )
        except MapGenerationError:
            continue
        require_car_spawn = not stage.endurance_stage
        car_reachable = validate_connectivity(
            blueprint.grid,
            fuel_mode=(
                stage.fuel_mode if not stage.endurance_stage else FuelMode.START_FULL
            ),
            require_player_exit_path=stage.endurance_stage,
            require_car_spawn=require_car_spawn,
        )
        if car_reachable is not None:
            blueprint.car_reachable_cells = car_reachable
            blueprint_data = blueprint
            break
    if blueprint_data is None:
        raise MapGenerationError(
            "Blueprint generation/connectivity validation failed after 20 attempts"
        )
    blueprint = blueprint_data.grid
    steel_cells_raw = blueprint_data.steel_cells
    car_reachable_cells = blueprint_data.car_reachable_cells

    steel_cells = (
        {(int(x), int(y)) for x, y in steel_cells_raw} if steel_enabled else set()
    )
    cell_size = stage.cell_size
    outer_wall_cells = {
        (x, y)
        for y, row in enumerate(blueprint)
        for x, ch in enumerate(row)
        if ch == "B"
    }
    wall_cells = {
        (x, y)
        for y, row in enumerate(blueprint)
        for x, ch in enumerate(row)
        if ch in {"B", "1", "R"}
    }
    steel_beam_cells: set[tuple[int, int]] = set()

    def _has_wall(nx: int, ny: int) -> bool:
        if nx < 0 or ny < 0 or nx >= stage.grid_cols or ny >= stage.grid_rows:
            return True
        return (nx, ny) in wall_cells

    outside_cells: set[tuple[int, int]] = set()
    moving_floor_cells: dict[tuple[int, int], MovingFloorDirection] = {}
    walkable_cells: list[tuple[int, int]] = []
    pitfall_cells: set[tuple[int, int]] = set()
    fire_floor_cells: set[tuple[int, int]] = set()
    metal_floor_cells: set[tuple[int, int]] = set()
    spiky_plant_cells: set[tuple[int, int]] = set()
    puddle_cells: set[tuple[int, int]] = set()
    player_cells: list[tuple[int, int]] = []
    car_cells: list[tuple[int, int]] = []
    fuel_cells: list[tuple[int, int]] = []
    empty_fuel_can_cells: list[tuple[int, int]] = []
    flashlight_cells: list[tuple[int, int]] = []
    shoes_cells: list[tuple[int, int]] = []
    interior_min_x = 2
    interior_max_x = stage.grid_cols - 3
    interior_min_y = 2
    interior_max_y = stage.grid_rows - 3
    bevel_corners: dict[tuple[int, int], tuple[bool, bool, bool, bool]] = {}
    palette = get_environment_palette(ambient_palette_key)
    rubble_ratio = max(0.0, min(1.0, stage.wall_rubble_ratio))

    def remove_steel_beam_cell(cell: tuple[int, int]) -> None:
        if cell in steel_beam_cells:
            steel_beam_cells.discard(cell)
        if (
            cell not in wall_cells
            and cell not in outer_wall_cells
            and cell not in pitfall_cells
            and cell not in walkable_cells
        ):
            walkable_cells.append(cell)

    def add_beam_to_groups(beam: SteelBeam, *, cell: tuple[int, int]) -> None:
        if beam._added_to_groups:
            return
        wall_group.add(beam)
        all_sprites.add(beam, layer=LAYER_WALLS)
        steel_beam_cells.add(cell)
        beam._added_to_groups = True

    def remove_wall_cell(cell: tuple[int, int], *, allow_walkable: bool = True) -> None:
        if cell in wall_cells:
            wall_cells.discard(cell)
            if allow_walkable and cell not in walkable_cells:
                walkable_cells.append(cell)
        outer_wall_cells.discard(cell)

    for y, row in enumerate(blueprint):
        if len(row) != stage.grid_cols:
            raise ValueError(
                f"Blueprint width mismatch at row {y}: {len(row)} != {stage.grid_cols}"
            )
        for x, ch in enumerate(row):
            cell_rect = _rect_for_cell(x, y, cell_size)
            cell_has_beam = steel_enabled and (x, y) in steel_cells
            if ch == "O":
                outside_cells.add((x, y))
                continue
            if ch == "B":
                draw_bottom_side = not _has_wall(x, y + 1)
                wall_cell = (x, y)
                wall = Wall(
                    cell_rect.x,
                    cell_rect.y,
                    cell_rect.width,
                    cell_rect.height,
                    health=OUTER_WALL_HEALTH,
                    palette=palette,
                    palette_category="outer_wall",
                    bevel_depth=0,
                    draw_bottom_side=draw_bottom_side,
                    on_destroy=(lambda _w, cell=wall_cell: remove_wall_cell(cell)),
                )
                wall_group.add(wall)
                all_sprites.add(wall, layer=LAYER_WALLS)
                continue
            if ch == "R":
                draw_bottom_side = not _has_wall(x, y + 1)
                bevel_mask = (
                    not _has_wall(x, y - 1)
                    and not _has_wall(x - 1, y)
                    and not _has_wall(x - 1, y - 1),
                    not _has_wall(x, y - 1)
                    and not _has_wall(x + 1, y)
                    and not _has_wall(x + 1, y - 1),
                    not _has_wall(x, y + 1)
                    and not _has_wall(x + 1, y)
                    and not _has_wall(x + 1, y + 1),
                    not _has_wall(x, y + 1)
                    and not _has_wall(x - 1, y)
                    and not _has_wall(x - 1, y + 1),
                )
                if any(bevel_mask):
                    bevel_corners[(x, y)] = bevel_mask
                wall_cell = (x, y)
                wall = ReinforcedWall(
                    cell_rect.x,
                    cell_rect.y,
                    cell_rect.width,
                    cell_rect.height,
                    health=OUTER_WALL_HEALTH,
                    palette=palette,
                    bevel_depth=INTERNAL_WALL_BEVEL_DEPTH,
                    bevel_mask=bevel_mask,
                    draw_bottom_side=draw_bottom_side,
                    on_destroy=(lambda _w, cell=wall_cell: remove_wall_cell(cell)),
                )
                wall_group.add(wall)
                all_sprites.add(wall, layer=LAYER_WALLS)
                continue
            if ch == "x":
                pitfall_cells.add((x, y))
                continue
            if ch == "F":
                fire_floor_cells.add((x, y))
                continue
            if ch == "m":
                metal_floor_cells.add((x, y))
                if not cell_has_beam:
                    walkable_cells.append((x, y))
                continue
            if ch == "h":
                spiky_plant_cells.add((x, y))
                if not cell_has_beam:
                    walkable_cells.append((x, y))
                continue
            if ch == "w":
                puddle_cells.add((x, y))
                if not cell_has_beam:
                    walkable_cells.append((x, y))
                continue
            if ch == "E":
                if not cell_has_beam:
                    walkable_cells.append((x, y))
            elif ch == "1":
                beam = None
                if cell_has_beam:
                    beam = SteelBeam(
                        cell_rect.x,
                        cell_rect.y,
                        cell_rect.width,
                        health=STEEL_BEAM_HEALTH,
                        palette=palette,
                        on_destroy=(
                            lambda _b, cell=(x, y): remove_steel_beam_cell(cell)
                        ),
                    )
                draw_bottom_side = not _has_wall(x, y + 1)
                bevel_mask = (
                    not _has_wall(x, y - 1)
                    and not _has_wall(x - 1, y)
                    and not _has_wall(x - 1, y - 1),
                    not _has_wall(x, y - 1)
                    and not _has_wall(x + 1, y)
                    and not _has_wall(x + 1, y - 1),
                    not _has_wall(x, y + 1)
                    and not _has_wall(x + 1, y)
                    and not _has_wall(x + 1, y + 1),
                    not _has_wall(x, y + 1)
                    and not _has_wall(x - 1, y)
                    and not _has_wall(x - 1, y + 1),
                )
                if any(bevel_mask):
                    bevel_corners[(x, y)] = bevel_mask
                wall_cell = (x, y)
                use_rubble = rubble_ratio > 0 and random.random() < rubble_ratio
                if use_rubble:
                    rotation_deg = (
                        RUBBLE_ROTATION_DEG
                        if random.random() < 0.5
                        else -RUBBLE_ROTATION_DEG
                    )
                    wall = RubbleWall(
                        cell_rect.x,
                        cell_rect.y,
                        cell_rect.width,
                        cell_rect.height,
                        health=INTERNAL_WALL_HEALTH,
                        palette=palette,
                        palette_category="inner_wall",
                        bevel_depth=INTERNAL_WALL_BEVEL_DEPTH,
                        rubble_rotation_deg=rotation_deg,
                        on_destroy=(
                            (
                                lambda _w, b=beam, cell=wall_cell: (
                                    remove_wall_cell(cell, allow_walkable=False),
                                    add_beam_to_groups(b, cell=cell),
                                )
                            )
                            if beam
                            else (lambda _w, cell=wall_cell: remove_wall_cell(cell))
                        ),
                    )
                else:
                    wall = Wall(
                        cell_rect.x,
                        cell_rect.y,
                        cell_rect.width,
                        cell_rect.height,
                        health=INTERNAL_WALL_HEALTH,
                        palette=palette,
                        palette_category="inner_wall",
                        bevel_mask=bevel_mask,
                        draw_bottom_side=draw_bottom_side,
                        on_destroy=(
                            (
                                lambda _w, b=beam, cell=wall_cell: (
                                    remove_wall_cell(cell, allow_walkable=False),
                                    add_beam_to_groups(b, cell=cell),
                                )
                            )
                            if beam
                            else (lambda _w, cell=wall_cell: remove_wall_cell(cell))
                        ),
                    )
                wall_group.add(wall)
                all_sprites.add(wall, layer=LAYER_WALLS)
            else:
                if not cell_has_beam:
                    walkable_cells.append((x, y))

            if ch in {"^", "v", "<", ">"}:
                if ch == "^":
                    moving_floor_cells[(x, y)] = MovingFloorDirection.UP
                elif ch == "v":
                    moving_floor_cells[(x, y)] = MovingFloorDirection.DOWN
                elif ch == "<":
                    moving_floor_cells[(x, y)] = MovingFloorDirection.LEFT
                else:
                    moving_floor_cells[(x, y)] = MovingFloorDirection.RIGHT
                if not cell_has_beam and (x, y) not in walkable_cells:
                    walkable_cells.append((x, y))
            if ch == "P":
                player_cells.append((x, y))
            if ch == "C":
                car_cells.append((x, y))
            if ch == "f":
                fuel_cells.append((x, y))
            if ch == "e":
                empty_fuel_can_cells.append((x, y))
            if ch == "l":
                flashlight_cells.append((x, y))
            if ch == "s":
                shoes_cells.append((x, y))

            if cell_has_beam and ch != "1":
                beam = SteelBeam(
                    cell_rect.x,
                    cell_rect.y,
                    cell_rect.width,
                    health=STEEL_BEAM_HEALTH,
                    palette=palette,
                    on_destroy=(lambda _b, cell=(x, y): remove_steel_beam_cell(cell)),
                )
                add_beam_to_groups(beam, cell=(x, y))

    layout = LevelLayout(
        field_rect=pygame.Rect(
            0,
            0,
            stage.grid_cols * cell_size,
            stage.grid_rows * cell_size,
        ),
        grid_cols=stage.grid_cols,
        grid_rows=stage.grid_rows,
        outside_cells=outside_cells,
        walkable_cells=[],
        outer_wall_cells=outer_wall_cells,
        wall_cells=wall_cells,
        steel_beam_cells=steel_beam_cells,
        pitfall_cells=pitfall_cells,
        car_walkable_cells=car_reachable_cells,
        car_spawn_cells=[],
        fall_spawn_cells=set(),
        spiky_plant_cells=spiky_plant_cells,
        fire_floor_cells=fire_floor_cells,
        metal_floor_cells=metal_floor_cells,
        zombie_contaminated_cells=set(),
        puddle_cells=puddle_cells,
        bevel_corners=bevel_corners,
        moving_floor_cells={},
    )
    if moving_floor_cells:
        pitfall_cells.difference_update(moving_floor_cells.keys())
        for cell in moving_floor_cells:
            if cell not in walkable_cells:
                walkable_cells.append(cell)
    if pitfall_cells:
        walkable_cells = [cell for cell in walkable_cells if cell not in pitfall_cells]
    if fire_floor_cells:
        walkable_cells = [cell for cell in walkable_cells if cell not in fire_floor_cells]
    layout.walkable_cells = walkable_cells
    layout.outer_wall_cells = outer_wall_cells
    layout.wall_cells = wall_cells
    layout.steel_beam_cells = steel_beam_cells
    layout.pitfall_cells = pitfall_cells
    layout.car_walkable_cells = car_reachable_cells
    layout.moving_floor_cells = moving_floor_cells
    fall_spawn_cells = _expand_zone_cells(
        stage.fall_spawn_zones,
        grid_cols=stage.grid_cols,
        grid_rows=stage.grid_rows,
    )
    walkable_set = set(walkable_cells)
    floor_ratio = max(0.0, min(1.0, stage.fall_spawn_cell_ratio))
    if floor_ratio > 0.0 and interior_min_x <= interior_max_x:
        for y in range(interior_min_y, interior_max_y + 1):
            for x in range(interior_min_x, interior_max_x + 1):
                cell = (x, y)
                if cell not in walkable_set:
                    continue
                if RNG.random() < floor_ratio:
                    fall_spawn_cells.add(cell)
        if not fall_spawn_cells:
            candidates = [
                cell
                for cell in walkable_set
                if (
                    interior_min_x <= cell[0] <= interior_max_x
                    and interior_min_y <= cell[1] <= interior_max_y
                )
            ]
            if candidates:
                fall_spawn_cells.add(RNG.choice(candidates))
    layout.fall_spawn_cells = fall_spawn_cells
    layout.bevel_corners = bevel_corners

    moving_floor_set = set(moving_floor_cells)
    item_spawn_cells = (
        [
            cell
            for cell in walkable_cells
            if cell not in moving_floor_set and cell not in fire_floor_cells
        ]
        if moving_floor_set
        else [cell for cell in walkable_cells if cell not in fire_floor_cells]
    )
    car_spawn_cells = (
        [cell for cell in car_reachable_cells if cell not in moving_floor_set]
        if moving_floor_set
        else list(car_reachable_cells)
    )
    layout.car_spawn_cells = car_spawn_cells

    return (
        layout,
        {
            "player_cells": player_cells,
            "car_cells": list(car_cells),
            "fuel_cells": list(fuel_cells),
            "empty_fuel_can_cells": list(empty_fuel_can_cells),
            "fuel_station_cells": list(fuel_cells),
            "flashlight_cells": list(flashlight_cells),
            "shoes_cells": list(shoes_cells),
            "spiky_plant_cells": list(spiky_plant_cells),
            "fire_floor_cells": list(fire_floor_cells),
            "metal_floor_cells": list(metal_floor_cells),
            "zombie_contaminated_cells": list(layout.zombie_contaminated_cells),
            "puddle_cells": list(puddle_cells),
            "walkable_cells": walkable_cells,
            "car_walkable_cells": list(car_reachable_cells),
            "item_spawn_cells": item_spawn_cells,
            "car_spawn_cells": car_spawn_cells,
        },
        wall_group,
        all_sprites,
        blueprint_data,
    )



from __future__ import annotations

import numpy as np  # type: ignore
import pygame


def build_grayscale_image(source: pygame.Surface) -> pygame.Surface:
    image = source.copy()
    image = image.convert_alpha()
    rgb = pygame.surfarray.pixels3d(image)
    alpha = pygame.surfarray.pixels_alpha(image)
    gray = (0.299 * rgb[:, :, 0] + 0.587 * rgb[:, :, 1] + 0.114 * rgb[:, :, 2]).astype(
        np.uint8
    )
    rgb[:, :, 0] = gray
    rgb[:, :, 1] = gray
    rgb[:, :, 2] = gray
    del rgb, alpha
    return image



from __future__ import annotations

import math
from enum import Enum
from typing import Protocol

import pygame

try:
    from typing import Self
except ImportError:  # pragma: no cover - Python 3.10 fallback
    from typing_extensions import Self

from ..entities_constants import (
    ZombieKind,
    ZOMBIE_DOG_ASSAULT_SPEED,
    ZOMBIE_DOG_CHARGE_COOLDOWN_MS_NIMBLE,
    ZOMBIE_DOG_CHARGE_COOLDOWN_MS_NORMAL,
    ZOMBIE_DOG_CHARGE_DISTANCE_NIMBLE,
    ZOMBIE_DOG_CHARGE_DISTANCE_NORMAL,
    ZOMBIE_DOG_CHARGE_OFFSET_MAX_NIMBLE,
    ZOMBIE_DOG_CHARGE_OFFSET_MAX_NORMAL,
    ZOMBIE_DOG_CHARGE_OFFSET_MIN_NIMBLE,
    ZOMBIE_DOG_CHARGE_OFFSET_MIN_NORMAL,
    ZOMBIE_CARBONIZE_DECAY_FRAMES,
    ZOMBIE_DOG_DECAY_DURATION_FRAMES,
    ZOMBIE_DOG_DECAY_MIN_SPEED_RATIO,
    ZOMBIE_DOG_LONG_AXIS_RATIO,
    ZOMBIE_DOG_PATROL_SPEED,
    ZOMBIE_DOG_PACK_CHASE_RANGE,
    ZOMBIE_DOG_SHORT_AXIS_RATIO,
    ZOMBIE_DOG_SIGHT_RANGE,
    ZOMBIE_DOG_WANDER_INTERVAL_MS,
    PATROL_BOT_ZOMBIE_DAMAGE,
    PATROL_BOT_ZOMBIE_DAMAGE_INTERVAL_FRAMES,
    PATROL_BOT_PARALYZE_MS,
    PATROL_BOT_PARALYZE_BLINK_MS,
    PATROL_BOT_PARALYZE_MARKER_COLOR,
    PUDDLE_SPEED_FACTOR,
    TRAPPED_ZOMBIE_SLOW_FACTOR,
    TRAPPED_ZOMBIE_REPEL_MAX_MULT,
    TRAPPED_ZOMBIE_REPEL_PER_STACK,
    TRAPPED_ZOMBIE_REPEL_RADIUS_CELLS,
    ZOMBIE_CONTAMINATED_SPEED_FACTOR,
    ZOMBIE_RADIUS,
    ZOMBIE_SEPARATION_DISTANCE,
)
from ..rng import get_rng
from ..surface_effects import SpikyPlantLike, is_in_contaminated_cell, is_in_puddle_cell
from ..render.entity_overlays import draw_paralyze_marker_overlay
from ..render_constants import ENTITY_SHADOW_RADIUS_MULT
from ..render_assets import (
    angle_bin_from_vector,
    build_zombie_dog_directional_surfaces,
)
from ..world_grid import apply_cell_edge_nudge
from .zombie import Zombie
from .movement import _circle_rect_collision
from .zombie_visuals import build_grayscale_image
from .zombie_vitals import ZombieVitals


RNG = get_rng()


class ZombieDogMode(Enum):
    WANDER = "wander"
    CHARGE = "charge"
    CHASE = "chase"


class ZombieDogVariant(str, Enum):
    NORMAL = "normal"
    NIMBLE = "nimble"


class MovementStrategy(Protocol):
    def __call__(
        self,
        zombie_dog: "ZombieDog",
        cell_size: int,
        layout,
        player_center: tuple[float, float],
        nearby_zombies: list[pygame.sprite.Sprite],
        footprints: list,
        *,
        now_ms: int,
    ) -> tuple[float, float]: ...


def _zombie_dog_default_movement(
    zombie_dog: "ZombieDog",
    cell_size: int,
    layout,
    player_center: tuple[float, float],
    nearby_zombies: list[pygame.sprite.Sprite],
    _footprints: list,
    *,
    now_ms: int,
) -> tuple[float, float]:
    def _pick_charge_target() -> tuple[float, float] | None:
        for _ in range(8):
            angle = RNG.uniform(0.0, math.tau)
            offset = RNG.uniform(
                zombie_dog.charge_offset_min, zombie_dog.charge_offset_max
            )
            tx = player_center[0] + math.cos(angle) * offset
            ty = player_center[1] + math.sin(angle) * offset
            if not (
                0 <= tx < layout.field_rect.width and 0 <= ty < layout.field_rect.height
            ):
                continue
            if cell_size > 0:
                target_cell = (int(tx // cell_size), int(ty // cell_size))
                if target_cell in layout.wall_cells:
                    continue
            # Avoid micro-charge that looks like jitter.
            if (
                math.hypot(tx - zombie_dog.x, ty - zombie_dog.y)
                < zombie_dog.speed_assault * 6.0
            ):
                continue
            return (tx, ty)
        return None

    nearby = list(nearby_zombies)
    chase_target = zombie_dog._nearest_zombie_target(nearby)
    if chase_target is not None:
        zombie_dog._set_mode(ZombieDogMode.CHASE)
    else:
        if zombie_dog.mode == ZombieDogMode.CHASE:
            zombie_dog._set_mode(ZombieDogMode.WANDER)
        chase_target = None

    if (
        zombie_dog.mode == ZombieDogMode.WANDER
        and zombie_dog._in_sight(player_center)
        and now_ms >= zombie_dog.next_charge_available_ms
    ):
        candidate = _pick_charge_target()
        if candidate is not None:
            zombie_dog._set_mode(ZombieDogMode.CHARGE)
            zombie_dog.wander_change_time = now_ms
            zombie_dog.charge_target = candidate
            dir_dx = candidate[0] - zombie_dog.x
            dir_dy = candidate[1] - zombie_dog.y
            dir_dist = math.hypot(dir_dx, dir_dy)
            if dir_dist > 1e-6:
                zombie_dog.charge_direction = (dir_dx / dir_dist, dir_dy / dir_dist)
            else:
                zombie_dog.charge_direction = None
            zombie_dog.charge_distance_remaining = zombie_dog.charge_distance_max
    elif zombie_dog.mode == ZombieDogMode.CHARGE and not zombie_dog._in_sight(
        player_center
    ):
        zombie_dog._set_mode(ZombieDogMode.WANDER)
        zombie_dog.next_charge_available_ms = now_ms + zombie_dog.charge_cooldown_ms
        zombie_dog.wander_change_time = now_ms
        zombie_dog.charge_direction = None

    if zombie_dog.mode == ZombieDogMode.WANDER:
        if now_ms - zombie_dog.wander_change_time > ZOMBIE_DOG_WANDER_INTERVAL_MS:
            zombie_dog.wander_change_time = now_ms
            zombie_dog.wander_angle = RNG.uniform(0.0, math.tau)
        move_x = math.cos(zombie_dog.wander_angle) * zombie_dog.speed_patrol
        move_y = math.sin(zombie_dog.wander_angle) * zombie_dog.speed_patrol
        if cell_size > 0 and layout.fire_floor_cells:
            next_cell = (
                int((zombie_dog.x + move_x) // cell_size),
                int((zombie_dog.y + move_y) // cell_size),
            )
            if next_cell in layout.fire_floor_cells:
                zombie_dog.wander_angle = (zombie_dog.wander_angle + math.pi) % math.tau
                move_x = math.cos(zombie_dog.wander_angle) * zombie_dog.speed_patrol
                move_y = math.sin(zombie_dog.wander_angle) * zombie_dog.speed_patrol
                retry_cell = (
                    int((zombie_dog.x + move_x) // cell_size),
                    int((zombie_dog.y + move_y) // cell_size),
                )
                if retry_cell in layout.fire_floor_cells:
                    return (0.0, 0.0)
        return (move_x, move_y)
    if zombie_dog.mode == ZombieDogMode.CHARGE:
        if zombie_dog.charge_target is None:
            zombie_dog._set_mode(ZombieDogMode.WANDER)
            zombie_dog.next_charge_available_ms = now_ms + zombie_dog.charge_cooldown_ms
            zombie_dog.wander_change_time = now_ms
            zombie_dog.charge_direction = None
            return (0.0, 0.0)
        if zombie_dog.charge_distance_remaining <= 0.0:
            zombie_dog._set_mode(ZombieDogMode.WANDER)
            zombie_dog.next_charge_available_ms = now_ms + zombie_dog.charge_cooldown_ms
            zombie_dog.wander_change_time = now_ms
            zombie_dog.charge_direction = None
            return (0.0, 0.0)
        charge_direction = zombie_dog.charge_direction
        if charge_direction is None:
            dx = zombie_dog.charge_target[0] - zombie_dog.x
            dy = zombie_dog.charge_target[1] - zombie_dog.y
            dist = math.hypot(dx, dy)
            if dist <= 1e-6:
                zombie_dog._set_mode(ZombieDogMode.WANDER)
                zombie_dog.next_charge_available_ms = (
                    now_ms + zombie_dog.charge_cooldown_ms
                )
                zombie_dog.wander_change_time = now_ms
                return (0.0, 0.0)
            charge_direction = (dx / dist, dy / dist)
            zombie_dog.charge_direction = charge_direction
        step = min(zombie_dog.speed_assault, zombie_dog.charge_distance_remaining)
        zombie_dog.charge_distance_remaining = max(
            0.0, zombie_dog.charge_distance_remaining - step
        )
        return (charge_direction[0] * step, charge_direction[1] * step)
    if zombie_dog.mode == ZombieDogMode.CHASE and chase_target is not None:
        dx = chase_target.rect.centerx - zombie_dog.x
        dy = chase_target.rect.centery - zombie_dog.y
        dist = math.hypot(dx, dy)
        if dist > 0:
            return (
                (dx / dist) * zombie_dog.speed_patrol,
                (dy / dist) * zombie_dog.speed_patrol,
            )
    zombie_dog._set_mode(ZombieDogMode.WANDER)
    zombie_dog.wander_angle = RNG.uniform(0.0, math.tau)
    zombie_dog.wander_change_time = now_ms
    return (
        math.cos(zombie_dog.wander_angle) * zombie_dog.speed_patrol,
        math.sin(zombie_dog.wander_angle) * zombie_dog.speed_patrol,
    )


class ZombieDog(pygame.sprite.Sprite):
    def __init__(
        self: Self,
        x: float,
        y: float,
        *,
        movement_strategy: MovementStrategy | None = None,
        variant: ZombieDogVariant = ZombieDogVariant.NORMAL,
    ) -> None:
        super().__init__()
        base_size = ZOMBIE_RADIUS * 2.0
        self.long_axis = base_size * ZOMBIE_DOG_LONG_AXIS_RATIO
        self.short_axis = base_size * ZOMBIE_DOG_SHORT_AXIS_RATIO
        self.radius = self.short_axis * 0.5
        self.speed_patrol = ZOMBIE_DOG_PATROL_SPEED
        self.speed_assault = ZOMBIE_DOG_ASSAULT_SPEED
        self.initial_speed_patrol = self.speed_patrol
        self.initial_speed_assault = self.speed_assault
        self.sight_range = ZOMBIE_DOG_SIGHT_RANGE
        self.mode = ZombieDogMode.WANDER
        self.charge_target: tuple[float, float] | None = None
        self.charge_direction: tuple[float, float] | None = None
        self.charge_distance_remaining = 0.0
        self.next_charge_available_ms = 0
        self.wander_angle = RNG.uniform(0.0, math.tau)
        self.wander_change_time = 0
        self.movement_strategy = movement_strategy or _zombie_dog_default_movement
        self.variant = (
            variant
            if isinstance(variant, ZombieDogVariant)
            else ZombieDogVariant(str(variant))
        )
        self.nimble_tail_side = RNG.choice([-1.0, 1.0])
        if self.variant == ZombieDogVariant.NIMBLE:
            self.charge_offset_min = ZOMBIE_DOG_CHARGE_OFFSET_MIN_NIMBLE
            self.charge_offset_max = ZOMBIE_DOG_CHARGE_OFFSET_MAX_NIMBLE
            self.charge_distance_max = ZOMBIE_DOG_CHARGE_DISTANCE_NIMBLE
            self.charge_cooldown_ms = ZOMBIE_DOG_CHARGE_COOLDOWN_MS_NIMBLE
        else:
            self.charge_offset_min = ZOMBIE_DOG_CHARGE_OFFSET_MIN_NORMAL
            self.charge_offset_max = ZOMBIE_DOG_CHARGE_OFFSET_MAX_NORMAL
            self.charge_distance_max = ZOMBIE_DOG_CHARGE_DISTANCE_NORMAL
            self.charge_cooldown_ms = ZOMBIE_DOG_CHARGE_COOLDOWN_MS_NORMAL
        self.kind = ZombieKind.DOG
        self.facing_bin = 0
        self.directional_images = build_zombie_dog_directional_surfaces(
            self.long_axis,
            self.short_axis,
        )
        if self.variant == ZombieDogVariant.NIMBLE:
            self.directional_images = self._build_nimble_directional_images(
                self.directional_images
            )
        self.image = self.directional_images[self.facing_bin]
        self.rect = self.image.get_rect(center=(x, y))
        self.x = float(self.rect.centerx)
        self.y = float(self.rect.centery)
        self.last_move_dx = 0.0
        self.last_move_dy = 0.0
        self.vitals = ZombieVitals(
            max_health=100,
            decay_duration_frames=ZOMBIE_DOG_DECAY_DURATION_FRAMES,
            decay_min_speed_ratio=ZOMBIE_DOG_DECAY_MIN_SPEED_RATIO,
            carbonize_decay_frames=ZOMBIE_CARBONIZE_DECAY_FRAMES,
            on_health_ratio=self._apply_speed_ratio,
            on_kill=self.kill,
            on_carbonize=self._apply_carbonize_visuals,
        )
        self.collision_radius = float(self.radius)
        self.shadow_radius = max(
            1, int(self.collision_radius * ENTITY_SHADOW_RADIUS_MULT)
        )
        self.shadow_offset_scale = 1.0

    def _build_nimble_directional_images(
        self: Self, base_images: list[pygame.Surface]
    ) -> list[pygame.Surface]:
        if not base_images:
            return base_images
        bins = max(1, len(base_images))
        tail_fill = (230, 40, 40)
        tail_radius = max(1, int(round(self.short_axis / 3.0)))
        rear_offset = max(1, int(round(self.long_axis * 0.35)))
        side_offset = max(1, int(round(tail_radius * 0.9)))
        side_sign = self.nimble_tail_side
        nimble_images: list[pygame.Surface] = []
        for idx, image in enumerate(base_images):
            surf = image.copy()
            center = pygame.Vector2(surf.get_rect().center)
            angle = (idx % bins) * (math.tau / bins)
            forward = pygame.Vector2(math.cos(angle), math.sin(angle))
            right = pygame.Vector2(-forward.y, forward.x)
            tail_center = (
                center - (forward * rear_offset) + (right * side_sign * side_offset)
            )
            tail_pos = (int(round(tail_center.x)), int(round(tail_center.y)))
            pygame.draw.circle(surf, tail_fill, tail_pos, tail_radius)
            nimble_images.append(surf)
        return nimble_images

    @property
    def max_health(self: Self) -> int:
        return self.vitals.max_health

    @property
    def health(self: Self) -> int:
        return self.vitals.health

    @property
    def decay_duration_frames(self: Self) -> float:
        return self.vitals.decay_duration_frames

    @property
    def carbonized(self: Self) -> bool:
        return self.vitals.carbonized

    @property
    def last_damage_ms(self: Self) -> int | None:
        return self.vitals.last_damage_ms

    @property
    def last_damage_source(self: Self) -> str | None:
        return self.vitals.last_damage_source

    @property
    def patrol_paralyze_until_ms(self: Self) -> int:
        return self.vitals.patrol_paralyze_until_ms

    @property
    def patrol_damage_frame_counter(self: Self) -> int:
        return self.vitals.patrol_damage_frame_counter

    def _apply_speed_ratio(self: Self, ratio: float) -> None:
        self.speed_patrol = self.initial_speed_patrol * ratio
        self.speed_assault = self.initial_speed_assault * ratio

    def _apply_carbonize_visuals(self: Self) -> None:
        self.image = build_grayscale_image(self.image)

    def _set_mode(self: Self, new_mode: ZombieDogMode) -> None:
        if self.mode == new_mode:
            return
        self.mode = new_mode

    def _set_facing_bin(self: Self, new_bin: int) -> None:
        if new_bin == self.facing_bin:
            return
        center = self.rect.center
        self.facing_bin = new_bin
        self.image = self.directional_images[self.facing_bin]
        self.rect = self.image.get_rect(center=center)

    def _update_facing_from_movement(self: Self, dx: float, dy: float) -> None:
        new_bin = angle_bin_from_vector(dx, dy)
        if new_bin is None:
            return
        self._set_facing_bin(new_bin)

    def _in_sight(self: Self, player_center: tuple[float, float]) -> bool:
        dx = player_center[0] - self.x
        dy = player_center[1] - self.y
        return dx * dx + dy * dy <= self.sight_range * self.sight_range

    def _nearest_zombie_target(
        self: Self, nearby_zombies: list[pygame.sprite.Sprite]
    ) -> pygame.sprite.Sprite | None:
        best: pygame.sprite.Sprite | None = None
        best_dist_sq = ZOMBIE_DOG_PACK_CHASE_RANGE * ZOMBIE_DOG_PACK_CHASE_RANGE
        for candidate in nearby_zombies:
            if not isinstance(candidate, Zombie):
                continue
            if candidate is self or not candidate.alive():
                continue
            if getattr(candidate, "is_trapped", False):
                continue
            if not hasattr(candidate, "carbonized"):
                continue
            if getattr(candidate, "carbonized", False):
                continue
            dx = candidate.rect.centerx - self.x
            dy = candidate.rect.centery - self.y
            dist_sq = dx * dx + dy * dy
            if dist_sq <= best_dist_sq:
                best = candidate
                best_dist_sq = dist_sq
        return best

    def _avoid_other_zombies(
        self: Self,
        move_x: float,
        move_y: float,
        zombies: list[pygame.sprite.Sprite],
    ) -> tuple[float, float]:
        """If another zombie is too close, steer directly away from the closest one."""
        next_x = self.x + move_x
        next_y = self.y + move_y

        closest: pygame.sprite.Sprite | None = None
        closest_dist_sq = ZOMBIE_SEPARATION_DISTANCE * ZOMBIE_SEPARATION_DISTANCE
        for other in zombies:
            if other is self or not other.alive():
                continue
            if getattr(other, "is_trapped", False):
                continue

            ox = other.x  # type: ignore[attr-defined]
            oy = other.y  # type: ignore[attr-defined]

            dx = ox - next_x
            dy = oy - next_y
            if (
                abs(dx) > ZOMBIE_SEPARATION_DISTANCE
                or abs(dy) > ZOMBIE_SEPARATION_DISTANCE
            ):
                continue
            dist_sq = dx * dx + dy * dy
            if dist_sq < closest_dist_sq:
                closest = other
                closest_dist_sq = dist_sq

        if closest is None:
            return move_x, move_y

        away_dx = next_x - closest.x  # type: ignore[attr-defined]
        away_dy = next_y - closest.y  # type: ignore[attr-defined]
        away_dist = math.hypot(away_dx, away_dy)
        if away_dist == 0:
            angle = RNG.uniform(0, 2 * math.pi)
            away_dx, away_dy = math.cos(angle), math.sin(angle)
            away_dist = 1

        speed = math.hypot(move_x, move_y)
        if speed <= 0:
            speed = self.speed_patrol
        move_x = (away_dx / away_dist) * speed
        move_y = (away_dy / away_dist) * speed
        return move_x, move_y

    def _slow_near_trapped_zombies(
        self: Self,
        move_x: float,
        move_y: float,
        zombies: list[pygame.sprite.Sprite],
    ) -> tuple[float, float]:
        next_x = self.x + move_x
        next_y = self.y + move_y
        for other in zombies:
            if other is self or not other.alive():
                continue
            if not getattr(other, "is_trapped", False):
                continue
            ox = float(getattr(other, "x", other.rect.centerx))
            oy = float(getattr(other, "y", other.rect.centery))
            other_radius = float(getattr(other, "collision_radius", 0.0))
            dx = ox - next_x
            dy = oy - next_y
            touch_radius = self.collision_radius + other_radius
            if dx * dx + dy * dy <= touch_radius * touch_radius:
                return (
                    move_x * TRAPPED_ZOMBIE_SLOW_FACTOR,
                    move_y * TRAPPED_ZOMBIE_SLOW_FACTOR,
                )
        return move_x, move_y

    def _repel_from_loaded_spiky_plants(
        self: Self,
        move_x: float,
        move_y: float,
        *,
        cell_size: int,
        spiky_plants: dict[tuple[int, int], SpikyPlantLike] | None,
        trapped_spiky_plant_counts: dict[tuple[int, int], int] | None,
    ) -> tuple[float, float]:
        if cell_size <= 0 or not spiky_plants or not trapped_spiky_plant_counts:
            return move_x, move_y
        next_x = self.x + move_x
        next_y = self.y + move_y
        effect_radius = max(
            self.collision_radius * 2.0,
            float(cell_size) * TRAPPED_ZOMBIE_REPEL_RADIUS_CELLS,
        )
        repel_x = 0.0
        repel_y = 0.0
        for cell, trapped_count in trapped_spiky_plant_counts.items():
            if trapped_count <= 0:
                continue
            hp = spiky_plants.get(cell)
            if hp is None or not hp.alive():
                continue
            dx = next_x - hp.x
            dy = next_y - hp.y
            dist = math.hypot(dx, dy)
            if dist > effect_radius:
                continue
            if dist <= 0.001:
                angle = RNG.uniform(0.0, math.tau)
                dx = math.cos(angle)
                dy = math.sin(angle)
                dist = 1.0
            falloff = 1.0 - (dist / effect_radius)
            count_mult = min(
                TRAPPED_ZOMBIE_REPEL_MAX_MULT,
                trapped_count * TRAPPED_ZOMBIE_REPEL_PER_STACK,
            )
            repel_mag = self.speed_patrol * count_mult * falloff
            repel_x += (dx / dist) * repel_mag
            repel_y += (dy / dist) * repel_mag
        return move_x + repel_x, move_y + repel_y

    def _apply_decay(self: Self) -> None:
        self.vitals.apply_decay()

    def _apply_wall_collision(
        self: Self,
        next_x: float,
        next_y: float,
        *,
        cell_size: int,
        layout,
    ) -> tuple[float, float, bool, bool]:
        if cell_size <= 0:
            return next_x, next_y, False, False

        blocked_cells = set(layout.wall_cells)
        blocked_cells.update(layout.outer_wall_cells)
        blocked_cells.update(layout.steel_beam_cells)

        def _collides(x: float, y: float) -> bool:
            radius = self.collision_radius
            min_cell_x = max(0, int((x - radius) // cell_size))
            max_cell_x = min(layout.grid_cols - 1, int((x + radius) // cell_size))
            min_cell_y = max(0, int((y - radius) // cell_size))
            max_cell_y = min(layout.grid_rows - 1, int((y + radius) // cell_size))
            for cy in range(min_cell_y, max_cell_y + 1):
                for cx in range(min_cell_x, max_cell_x + 1):
                    if (cx, cy) not in blocked_cells:
                        continue
                    rect = pygame.Rect(
                        cx * cell_size,
                        cy * cell_size,
                        cell_size,
                        cell_size,
                    )
                    if _circle_rect_collision((x, y), radius, rect):
                        return True
            return False

        hit_x = False
        hit_y = False
        final_x = self.x
        final_y = self.y
        if next_x != self.x:
            if _collides(next_x, final_y):
                hit_x = True
                next_x = self.x
            final_x = next_x
        if next_y != self.y:
            if _collides(final_x, next_y):
                hit_y = True
                next_y = self.y
            final_y = next_y
        return final_x, final_y, hit_x, hit_y

    def _apply_paralyze_overlay(self: Self, now_ms: int) -> None:
        base_surface = self.directional_images[self.facing_bin]
        image = base_surface.copy()
        center = image.get_rect().center
        marker_size = max(6, int(self.short_axis * 0.8))
        draw_paralyze_marker_overlay(
            surface_out=image,
            now_ms=now_ms,
            blink_ms=PATROL_BOT_PARALYZE_BLINK_MS,
            center=center,
            size=marker_size,
            color=PATROL_BOT_PARALYZE_MARKER_COLOR,
            offset=int(self.short_axis * 0.4),
            width=2,
        )
        self.image = image

    def update(
        self: Self,
        player_center: tuple[float, float],
        walls: list[pygame.sprite.Sprite],
        nearby_zombies: list[pygame.sprite.Sprite],
        electrified_cells: set[tuple[int, int]] | None = None,
        footprints: list | None = None,
        *,
        cell_size: int,
        layout,
        now_ms: int,
        drift: tuple[float, float] = (0.0, 0.0),
        spiky_plants: dict[tuple[int, int], SpikyPlantLike] | None = None,
        trapped_spiky_plant_counts: dict[tuple[int, int], int] | None = None,
    ) -> None:
        if self.vitals.carbonized:
            self._apply_decay()
            return
        self._apply_decay()
        if not self.alive():
            return

        _ = nearby_zombies, footprints, walls
        now = now_ms
        drift_x, drift_y = drift
        on_electrified_floor = False
        if cell_size > 0 and electrified_cells:
            current_cell = (int(self.x // cell_size), int(self.y // cell_size))
            on_electrified_floor = current_cell in electrified_cells
        if self.vitals.update_patrol_floor_paralyze(
            on_electrified_floor=on_electrified_floor,
            now_ms=now,
            paralyze_duration_ms=PATROL_BOT_PARALYZE_MS,
            damage_interval_frames=PATROL_BOT_ZOMBIE_DAMAGE_INTERVAL_FRAMES,
            damage_amount=PATROL_BOT_ZOMBIE_DAMAGE,
            apply_damage=lambda amount: self.take_damage(amount, now_ms=now),
        ):
            self.last_move_dx = 0.0
            self.last_move_dy = 0.0
            self._apply_paralyze_overlay(now)
            return
        level_width = layout.field_rect.width
        level_height = layout.field_rect.height

        move_x, move_y = self.movement_strategy(
            self,
            cell_size,
            layout,
            player_center,
            list(nearby_zombies),
            footprints or [],
            now_ms=now,
        )

        move_x += drift_x
        move_y += drift_y

        # Puddle slow-down
        if is_in_puddle_cell(
            self.x,
            self.y,
            cell_size=cell_size,
            puddle_cells=layout.puddle_cells,
        ):
            move_x *= PUDDLE_SPEED_FACTOR
            move_y *= PUDDLE_SPEED_FACTOR
        if is_in_contaminated_cell(
            self.x,
            self.y,
            cell_size=cell_size,
            contaminated_cells=layout.zombie_contaminated_cells,
        ):
            move_x *= ZOMBIE_CONTAMINATED_SPEED_FACTOR
            move_y *= ZOMBIE_CONTAMINATED_SPEED_FACTOR
        if nearby_zombies:
            move_x, move_y = self._slow_near_trapped_zombies(
                move_x, move_y, list(nearby_zombies)
            )
        move_x, move_y = self._repel_from_loaded_spiky_plants(
            move_x,
            move_y,
            cell_size=cell_size,
            spiky_plants=spiky_plants,
            trapped_spiky_plant_counts=trapped_spiky_plant_counts,
        )

        if nearby_zombies:
            move_x, move_y = self._avoid_other_zombies(
                move_x, move_y, list(nearby_zombies)
            )
        move_x, move_y = apply_cell_edge_nudge(
            self.x,
            self.y,
            move_x,
            move_y,
            layout=layout,
            cell_size=cell_size,
        )
        self._update_facing_from_movement(move_x, move_y)
        if self.patrol_paralyze_until_ms > now:
            self._apply_paralyze_overlay(now)
        else:
            self.image = self.directional_images[self.facing_bin]
        self.last_move_dx = move_x
        self.last_move_dy = move_y

        next_x = self.x + move_x
        next_y = self.y + move_y
        final_x, final_y, hit_x, hit_y = self._apply_wall_collision(
            next_x,
            next_y,
            cell_size=cell_size,
            layout=layout,
        )

        if self.mode == ZombieDogMode.WANDER and (hit_x or hit_y):
            self.wander_angle = (self.wander_angle + math.pi) % math.tau
        elif self.mode in (ZombieDogMode.CHARGE, ZombieDogMode.CHASE) and (
            hit_x or hit_y
        ):
            was_charge = self.mode == ZombieDogMode.CHARGE
            self._set_mode(ZombieDogMode.WANDER)
            if was_charge:
                self.next_charge_available_ms = now + self.charge_cooldown_ms
                self.charge_direction = None
            self.wander_angle = RNG.uniform(0.0, math.tau)
            self.wander_change_time = now

        if (
            self.mode == ZombieDogMode.WANDER
            and self.last_move_dx == 0.0
            and self.last_move_dy == 0.0
        ):
            self.wander_angle = RNG.uniform(0.0, math.tau)
            self.wander_change_time = now
        if final_x == self.x and final_y == self.y:
            self.wander_angle = RNG.uniform(0.0, math.tau)
            try_dx = math.cos(self.wander_angle) * self.speed_patrol
            try_dy = math.sin(self.wander_angle) * self.speed_patrol
            try_dx, try_dy = apply_cell_edge_nudge(
                self.x,
                self.y,
                try_dx,
                try_dy,
                layout=layout,
                cell_size=cell_size,
            )
            retry_x, retry_y, _, _ = self._apply_wall_collision(
                self.x + try_dx,
                self.y + try_dy,
                cell_size=cell_size,
                layout=layout,
            )
            if retry_x != self.x or retry_y != self.y:
                final_x = retry_x
                final_y = retry_y

        if not (0 <= final_x < level_width and 0 <= final_y < level_height):
            self.kill()
            return

        self.x = final_x
        self.y = final_y
        self.rect.center = (int(self.x), int(self.y))

    def carbonize(self: Self) -> None:
        self.vitals.carbonize()

    def take_damage(self: Self, amount: int, *, now_ms: int) -> None:
        if amount <= 0 or not self.alive():
            return
        self.vitals.take_damage(amount, now_ms=now_ms)



"""Render-related constants and assets."""

from __future__ import annotations

import math
from dataclasses import dataclass

from .entities_constants import FOV_RADIUS, PLAYER_RADIUS
from .screen_constants import SCREEN_HEIGHT, SCREEN_WIDTH, STATUS_BAR_HEIGHT

# Keep constants grouped by subsystem so new values have a predictable home.

# --- Entity colors ---
HUMANOID_OUTLINE_COLOR = (0, 80, 200)
HUMANOID_OUTLINE_WIDTH = 1
BUDDY_COLOR = (0, 155, 140)
SURVIVOR_COLOR = (198, 198, 198)
ZOMBIE_BODY_COLOR = (180, 0, 0)
ZOMBIE_OUTLINE_COLOR = (255, 60, 60)
ZOMBIE_NOSE_COLOR = (255, 80, 80)
TRAPPED_OUTLINE_COLOR = (172, 0, 181)
PATROL_BOT_BODY_COLOR = (200, 200, 200)
PATROL_BOT_OUTLINE_COLOR = (45, 40, 50)
PATROL_BOT_ARROW_COLOR = (140, 70, 200)
SPIKY_PLANT_BODY_COLOR = (30, 120, 30)
SPIKY_PLANT_SPIKE_COLOR = (61, 196, 115)

# --- Tile/effect colors ---
MOVING_FLOOR_TILE_COLOR = (90, 90, 90)
MOVING_FLOOR_LINE_COLOR = (130, 130, 130)
MOVING_FLOOR_BORDER_COLOR = (10, 10, 10)
PUDDLE_TILE_COLOR = (30, 60, 100)
METAL_FLOOR_BASE_COLOR = (52, 56, 62)
METAL_FLOOR_LINE_COLOR = (78, 84, 94)
METAL_FLOOR_HIGHLIGHT_COLOR = (104, 112, 126)
METAL_FLOOR_HAIRLINE_COLOR = (122, 130, 144)
FIRE_FLOOR_BASE_COLOR = (88, 40, 22)
FIRE_FLOOR_FLAME_COLORS = (
    (214, 126, 48),
    (188, 92, 36),
    (160, 62, 26),
)
FIRE_GRATE_BASE_COLOR = (92, 96, 104)
FIRE_GRATE_EDGE_COLOR = (58, 62, 70)
FIRE_GLASS_BASE_COLOR = (18, 22, 28)
FIRE_GLASS_GRID_COLOR = (64, 72, 86)
FIRE_GLASS_HIGHLIGHT_COLOR = (125, 140, 165)
FALLING_ZOMBIE_COLOR = (45, 45, 45)
FALLING_WHIRLWIND_COLOR = (200, 200, 200, 120)
FALLING_DUST_COLOR = (70, 70, 70, 130)

# --- Overview colors ---
MOVING_FLOOR_OVERVIEW_COLOR = (50, 50, 50)
LINEFORMER_MARKER_OVERVIEW_COLOR = (150, 50, 50)
TRAPPED_ZOMBIE_OVERVIEW_COLOR = (110, 40, 140)
PUDDLE_OVERVIEW_COLOR = (60, 120, 200)

# --- Gameplay/UI tuning ---
ANGLE_BINS = 16
HAND_SPREAD_RAD = math.radians(75)
GAMEPLAY_FONT_SIZE = 11
HUD_ICON_SIZE = 12
FADE_IN_DURATION_MS = 900
TIMED_MESSAGE_LEFT_X = 20
TIMED_MESSAGE_TOP_Y = 48
TIMED_MESSAGE_BAND_ALPHA = 80

# --- Fog/FOV tuning ---
FOG_RADIUS_SCALE = 1.2
FOG_EDGE_SOFTNESS_PX = 6
FOG_LAYER_AA_SCALE = 4

FLASHLIGHT_FOG_SCALE_ONE = FOG_RADIUS_SCALE + 0.3
FLASHLIGHT_FOG_SCALE_TWO = FOG_RADIUS_SCALE + 0.6
_FLASHLIGHT_HATCH_EXTRA_SCALE = 0.12
FOG_HATCH_SOFTEN_SCALE = 1.0
FOG_HATCH_DENSITY_RAMPS: list[tuple[float, float]] = [
    (0.5, 0.5),
    (0.97, 0.3),
]

# --- Footprint tuning ---
FOOTPRINT_RADIUS = 2
FOOTPRINT_OVERVIEW_RADIUS = 3
FOOTPRINT_COLOR = (110, 200, 255)
FOOTPRINT_LIFETIME_MS = 135000
FOOTPRINT_MIN_FADE = 0.3

# --- Shadow tuning ---
SHADOW_OVERSAMPLE = 2
SHADOW_STEPS = 10
SHADOW_MIN_RATIO = 0.0
SHADOW_RADIUS_RATIO = 0.3
ENTITY_SHADOW_RADIUS_MULT = 1.8
ENTITY_SHADOW_ALPHA = 48
ENTITY_SHADOW_EDGE_SOFTNESS = 0.32
PLAYER_SHADOW_RADIUS_MULT = 1.2

# --- Pitfall rendering ---
PITFALL_ABYSS_COLOR = (21, 20, 20)
PITFALL_SHADOW_RIM_COLOR = (38, 34, 34)
PITFALL_SHADOW_WIDTH = 6
PITFALL_EDGE_METAL_COLOR = (110, 110, 115)
PITFALL_EDGE_STRIPE_COLOR = (75, 75, 80)
PITFALL_EDGE_STRIPE_SPACING = 6
PITFALL_EDGE_DEPTH_OFFSET = 3


@dataclass(frozen=True)
class RenderAssets:
    screen_width: int
    screen_height: int
    status_bar_height: int
    player_radius: int
    fov_radius: int
    fog_radius_scale: float
    footprint_radius: int
    footprint_overview_radius: int
    footprint_lifetime_ms: int
    footprint_min_fade: float
    internal_wall_grid_snap: int
    flashlight_hatch_extra_scale: float
    fog_hatch_soften_scale: float
    fog_hatch_density_ramps: list[tuple[float, float]]
    fog_edge_softness_px: int
    fog_layer_aa_scale: int


def build_render_assets(cell_size: int) -> RenderAssets:
    return RenderAssets(
        screen_width=SCREEN_WIDTH,
        screen_height=SCREEN_HEIGHT,
        status_bar_height=STATUS_BAR_HEIGHT,
        player_radius=PLAYER_RADIUS,
        fov_radius=FOV_RADIUS,
        fog_radius_scale=FOG_RADIUS_SCALE,
        footprint_radius=FOOTPRINT_RADIUS,
        footprint_overview_radius=FOOTPRINT_OVERVIEW_RADIUS,
        footprint_lifetime_ms=FOOTPRINT_LIFETIME_MS,
        footprint_min_fade=FOOTPRINT_MIN_FADE,
        internal_wall_grid_snap=cell_size,
        flashlight_hatch_extra_scale=_FLASHLIGHT_HATCH_EXTRA_SCALE,
        fog_hatch_soften_scale=FOG_HATCH_SOFTEN_SCALE,
        fog_hatch_density_ramps=FOG_HATCH_DENSITY_RAMPS,
        fog_edge_softness_px=FOG_EDGE_SOFTNESS_PX,
        fog_layer_aa_scale=FOG_LAYER_AA_SCALE,
    )


__all__ = [
    # Entity colors
    "HUMANOID_OUTLINE_COLOR",
    "HUMANOID_OUTLINE_WIDTH",
    "BUDDY_COLOR",
    "SURVIVOR_COLOR",
    "ZOMBIE_BODY_COLOR",
    "ZOMBIE_OUTLINE_COLOR",
    "ZOMBIE_NOSE_COLOR",
    "TRAPPED_OUTLINE_COLOR",
    "PATROL_BOT_BODY_COLOR",
    "PATROL_BOT_OUTLINE_COLOR",
    "PATROL_BOT_ARROW_COLOR",
    "SPIKY_PLANT_BODY_COLOR",
    "SPIKY_PLANT_SPIKE_COLOR",
    # Tile/effect colors
    "FALLING_ZOMBIE_COLOR",
    "FALLING_WHIRLWIND_COLOR",
    "FALLING_DUST_COLOR",
    "MOVING_FLOOR_TILE_COLOR",
    "MOVING_FLOOR_LINE_COLOR",
    "MOVING_FLOOR_BORDER_COLOR",
    "PUDDLE_TILE_COLOR",
    "METAL_FLOOR_BASE_COLOR",
    "METAL_FLOOR_LINE_COLOR",
    "METAL_FLOOR_HIGHLIGHT_COLOR",
    "METAL_FLOOR_HAIRLINE_COLOR",
    "FIRE_FLOOR_BASE_COLOR",
    "FIRE_FLOOR_FLAME_COLORS",
    # Overview colors
    "MOVING_FLOOR_OVERVIEW_COLOR",
    "LINEFORMER_MARKER_OVERVIEW_COLOR",
    "TRAPPED_ZOMBIE_OVERVIEW_COLOR",
    "PUDDLE_OVERVIEW_COLOR",
    # Gameplay/UI tuning
    "ANGLE_BINS",
    "HAND_SPREAD_RAD",
    "GAMEPLAY_FONT_SIZE",
    "HUD_ICON_SIZE",
    "FADE_IN_DURATION_MS",
    "TIMED_MESSAGE_LEFT_X",
    "TIMED_MESSAGE_TOP_Y",
    "TIMED_MESSAGE_BAND_ALPHA",
    # Fog/FOV tuning
    "FOG_RADIUS_SCALE",
    "FOG_EDGE_SOFTNESS_PX",
    "FOG_LAYER_AA_SCALE",
    "FLASHLIGHT_FOG_SCALE_ONE",
    "FLASHLIGHT_FOG_SCALE_TWO",
    "FOG_HATCH_SOFTEN_SCALE",
    "FOG_HATCH_DENSITY_RAMPS",
    # Shadow tuning
    "SHADOW_OVERSAMPLE",
    "SHADOW_STEPS",
    "SHADOW_MIN_RATIO",
    "SHADOW_RADIUS_RATIO",
    "ENTITY_SHADOW_RADIUS_MULT",
    "ENTITY_SHADOW_ALPHA",
    "ENTITY_SHADOW_EDGE_SOFTNESS",
    "PLAYER_SHADOW_RADIUS_MULT",
    # Pitfall rendering
    "PITFALL_ABYSS_COLOR",
    "PITFALL_SHADOW_RIM_COLOR",
    "PITFALL_SHADOW_WIDTH",
    "PITFALL_EDGE_METAL_COLOR",
    "PITFALL_EDGE_STRIPE_COLOR",
    "PITFALL_EDGE_STRIPE_SPACING",
    "PITFALL_EDGE_DEPTH_OFFSET",
    # Data/builder API
    "RenderAssets",
    "build_render_assets",
]



from __future__ import annotations

import math
from pathlib import Path

import pygame

from .colors import STUDIO_AMBIENT_PALETTE_KEY, get_environment_palette
from .entities import (
    Car,
    EmptyFuelCan,
    Flashlight,
    FuelCan,
    FuelStation,
    PatrolBot,
    Player,
    ReinforcedWall,
    RubbleWall,
    Shoes,
    SteelBeam,
    Survivor,
    Wall,
    Zombie,
    ZombieDog,
    SpikyPlant,
)
from .entities_constants import (
    INTERNAL_WALL_BEVEL_DEPTH,
    STEEL_BEAM_HEALTH,
    ZOMBIE_RADIUS,
    MovingFloorDirection,
    ZombieKind,
)
from .gameplay.state import initialize_game_state
from .level_constants import DEFAULT_CELL_SIZE
from .models import FallingEntity, FuelProgress, Stage
from .render.core import _draw_entities, _draw_falling_fx, _draw_play_area
from .render_constants import build_render_assets
from .render_constants import ENTITY_SHADOW_ALPHA, ENTITY_SHADOW_EDGE_SOFTNESS
from .render.shadows import _get_shadow_layer, draw_single_entity_shadow_by_mode
from .screen_constants import SCREEN_HEIGHT, SCREEN_WIDTH

__all__ = ["export_images"]


def _ensure_pygame_ready() -> None:
    if not pygame.get_init():
        pygame.init()
    if not pygame.display.get_init():
        pygame.display.init()
    if pygame.display.get_surface() is None:
        flags = pygame.HIDDEN if hasattr(pygame, "HIDDEN") else 0
        pygame.display.set_mode((1, 1), flags=flags)


def _save_surface(surface: pygame.Surface, path: Path, *, scale: int = 1) -> None:
    if scale != 1:
        width, height = surface.get_size()
        surface = pygame.transform.scale(surface, (width * scale, height * scale))
    path.parent.mkdir(parents=True, exist_ok=True)
    pygame.image.save(surface, str(path))


def _studio_grid_size(cell_size: int) -> tuple[int, int]:
    cols = max(3, math.ceil(SCREEN_WIDTH / cell_size) + 2)
    rows = max(3, math.ceil(SCREEN_HEIGHT / cell_size) + 2)
    return cols, rows


def _build_studio_stage(cell_size: int) -> Stage:
    cols, rows = _studio_grid_size(cell_size)
    return Stage(
        id="studio",
        name_key="studio",
        description_key="studio",
        available=False,
        cell_size=cell_size,
        grid_cols=cols,
        grid_rows=rows,
        zombie_normal_ratio=1.0,
    )


def _build_studio_game_data(cell_size: int):
    stage = _build_studio_stage(cell_size)
    game_data = initialize_game_state(stage)
    state = game_data.state
    state.ambient_palette_key = STUDIO_AMBIENT_PALETTE_KEY
    state.fuel_progress = FuelProgress.NONE
    state.flashlight_count = 0
    state.shoes_count = 0
    state.timed_message = None
    state.fade_in_started_at_ms = None
    state.show_fps = False
    state.debug_mode = False
    state.clock.elapsed_ms = 0
    state.falling_zombies = []
    state.dust_rings = []
    state.decay_effects = []
    state.footprints = []
    return game_data


def _center_camera(camera: object, target_rect: pygame.Rect) -> None:
    dummy = pygame.sprite.Sprite()
    dummy.rect = target_rect.copy()
    camera.update(dummy)


def _render_studio_snapshot(
    *,
    cell_size: int,
    target_rect: pygame.Rect,
    sprites: list[pygame.sprite.Sprite] | None = None,
    pitfall_cells: set[tuple[int, int]] | None = None,
    fire_floor_cells: set[tuple[int, int]] | None = None,
    fall_spawn_cells: set[tuple[int, int]] | None = None,
    moving_floor_cells: dict[tuple[int, int], MovingFloorDirection] | None = None,
    puddle_cells: set[tuple[int, int]] | None = None,
    falling_zombies: list[FallingEntity] | None = None,
    enable_shadows: bool = False,
    ambient_palette_key: str | None = STUDIO_AMBIENT_PALETTE_KEY,
) -> pygame.Surface:
    game_data = _build_studio_game_data(cell_size)
    game_data.state.ambient_palette_key = ambient_palette_key
    assets = build_render_assets(cell_size)
    screen = pygame.Surface(
        (assets.screen_width, assets.screen_height), pygame.SRCALPHA
    )

    layout = game_data.layout
    layout.pitfall_cells = pitfall_cells or set()
    layout.fire_floor_cells = fire_floor_cells or set()
    layout.metal_floor_cells = set()
    layout.fall_spawn_cells = fall_spawn_cells or set()
    layout.moving_floor_cells = moving_floor_cells or {}
    layout.puddle_cells = puddle_cells or set()

    sprites = sprites or []
    player = None
    for sprite in sprites:
        if isinstance(sprite, Player):
            player = sprite
        game_data.groups.all_sprites.add(sprite)
    if player is None:
        player = Player(-cell_size, -cell_size)
    game_data.player = player

    padding_x = max(1, int(round(target_rect.width * 0.3)))
    padding_y = max(1, int(round(target_rect.height * 0.3)))
    framing_rect = target_rect.inflate(padding_x, padding_y)

    _center_camera(game_data.camera, framing_rect)

    palette = get_environment_palette(game_data.state.ambient_palette_key)
    screen.fill(palette.outside)
    _draw_play_area(
        screen,
        game_data.camera.apply_rect,
        pygame.Rect(
            -game_data.camera.camera.x,
            -game_data.camera.camera.y,
            assets.screen_width,
            assets.screen_height,
        ),
        assets,
        palette,
        layout.field_rect,
        layout.outside_cells,
        layout.fall_spawn_cells,
        layout.pitfall_cells,
        layout.fire_floor_cells,
        layout.metal_floor_cells,
        layout.puddle_cells,
        layout.moving_floor_cells,
        set(),
        game_data.cell_size,
        elapsed_ms=int(game_data.state.clock.elapsed_ms),
    )
    if enable_shadows:
        shadow_layer = _get_shadow_layer(screen.get_size())
        shadow_layer.fill((0, 0, 0, 0))
        light_offset = max(4, int(min(target_rect.width, target_rect.height) * 0.14))
        light_source_pos = (
            float(target_rect.centerx - light_offset),
            float(target_rect.centery + light_offset),
        )
        shadow_alpha = max(1, int(ENTITY_SHADOW_ALPHA * 2.2))
        drew_shadow = False
        for sprite in sprites:
            if not sprite.alive():
                continue
            shadow_radius_raw = getattr(sprite, "shadow_radius", None)
            if shadow_radius_raw is not None:
                shadow_radius = max(0, int(shadow_radius_raw))
            else:
                collision_radius = getattr(sprite, "collision_radius", None)
                if collision_radius is not None:
                    shadow_radius = max(1, int(float(collision_radius) * 1.2))
                else:
                    shadow_radius = max(
                        1, int(min(sprite.rect.width, sprite.rect.height) * 0.5 * 1.2)
                    )
            offset_scale = float(getattr(sprite, "shadow_offset_scale", 1.0))
            drew_shadow |= draw_single_entity_shadow_by_mode(
                shadow_layer,
                game_data.camera.apply_rect,
                entity=sprite,
                dawn_shadow_mode=False,
                light_source_pos=light_source_pos,
                outside_cells=layout.outside_cells,
                cell_size=game_data.cell_size,
                shadow_radius=shadow_radius,
                alpha=shadow_alpha,
                edge_softness=ENTITY_SHADOW_EDGE_SOFTNESS,
                offset_scale=offset_scale,
            )
        if drew_shadow:
            screen.blit(shadow_layer, (0, 0))
    if falling_zombies:
        game_data.state.falling_zombies = falling_zombies
        _draw_falling_fx(
            screen,
            game_data.camera.apply_rect,
            game_data.state.falling_zombies,
            game_data.state.flashlight_count,
            game_data.state.dust_rings,
            int(game_data.state.clock.elapsed_ms),
        )
    _draw_entities(
        screen,
        [
            (entity, game_data.camera.apply_rect(entity.rect))
            for entity in game_data.groups.all_sprites
        ],
        player,
        has_fuel=(game_data.state.fuel_progress == FuelProgress.FULL_CAN),
        has_empty_fuel_can=(game_data.state.fuel_progress == FuelProgress.EMPTY_CAN),
        show_fuel_indicator=False,
    )

    screen_rect = game_data.camera.apply_rect(framing_rect)
    return screen.subsurface(screen_rect).copy()


def export_images(
    output_dir: Path, *, cell_size: int = DEFAULT_CELL_SIZE, output_scale: int = 4
) -> list[Path]:
    _ensure_pygame_ready()

    saved: list[Path] = []
    out = Path(output_dir)

    cols, rows = _studio_grid_size(cell_size)
    center_x = (cols * cell_size) // 2
    center_y = (rows * cell_size) // 2
    cell_x = cols // 2
    cell_y = rows // 2

    player = Player(center_x, center_y)
    player_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=player.rect,
        sprites=[player],
        enable_shadows=True,
    )
    player_path = out / "player.png"
    _save_surface(player_surface, player_path, scale=output_scale)
    saved.append(player_path)

    zombie = Zombie(center_x, center_y, kind=ZombieKind.NORMAL)
    zombie_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=zombie.rect,
        sprites=[zombie],
        enable_shadows=True,
    )
    zombie_normal_path = out / "zombie-normal.png"
    _save_surface(zombie_surface, zombie_normal_path, scale=output_scale)
    saved.append(zombie_normal_path)

    tracker = Zombie(center_x, center_y, kind=ZombieKind.TRACKER)
    tracker._apply_render_overlays()
    tracker_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=tracker.rect,
        sprites=[tracker],
        enable_shadows=True,
    )
    tracker_path = out / "zombie-tracker.png"
    _save_surface(tracker_surface, tracker_path, scale=output_scale)
    saved.append(tracker_path)

    wall_hugging = Zombie(center_x, center_y, kind=ZombieKind.WALL_HUGGER)
    wall_hugging.wall_hug_side = 1.0
    wall_hugging.wall_hug_last_side_has_wall = True
    wall_hugging._apply_render_overlays()
    wall_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=wall_hugging.rect,
        sprites=[wall_hugging],
        enable_shadows=True,
    )
    wall_path = out / "zombie-wall.png"
    _save_surface(wall_surface, wall_path, scale=output_scale)
    saved.append(wall_path)

    lineformer = Zombie(center_x, center_y, kind=ZombieKind.LINEFORMER)
    lineformer.lineformer_target_pos = (center_x + cell_size, center_y)
    lineformer._apply_render_overlays()
    lineformer_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=lineformer.rect,
        sprites=[lineformer],
        enable_shadows=True,
    )
    lineformer_path = out / "zombie-lineformer.png"
    _save_surface(lineformer_surface, lineformer_path, scale=output_scale)
    saved.append(lineformer_path)

    solitary = Zombie(center_x, center_y, kind=ZombieKind.SOLITARY)
    solitary._apply_render_overlays()
    solitary_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=solitary.rect,
        sprites=[solitary],
        enable_shadows=True,
    )
    solitary_path = out / "zombie-solitary.png"
    _save_surface(solitary_surface, solitary_path, scale=output_scale)
    saved.append(solitary_path)

    zombie_dog = ZombieDog(center_x, center_y)
    zombie_dog_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=zombie_dog.rect,
        sprites=[zombie_dog],
        enable_shadows=True,
    )
    zombie_dog_path = out / "zombie-dog.png"
    _save_surface(zombie_dog_surface, zombie_dog_path, scale=output_scale)
    saved.append(zombie_dog_path)

    zombie_dog_nimble = ZombieDog(center_x, center_y, variant="nimble")
    zombie_dog_nimble_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=zombie_dog_nimble.rect,
        sprites=[zombie_dog_nimble],
        enable_shadows=True,
    )
    zombie_dog_nimble_path = out / "zombie-dog-nimble.png"
    _save_surface(zombie_dog_nimble_surface, zombie_dog_nimble_path, scale=output_scale)
    saved.append(zombie_dog_nimble_path)

    buddy = Survivor(center_x, center_y, is_buddy=True)
    buddy_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=buddy.rect,
        sprites=[buddy],
        enable_shadows=True,
    )
    buddy_path = out / "buddy.png"
    _save_surface(buddy_surface, buddy_path, scale=output_scale)
    saved.append(buddy_path)

    survivor = Survivor(center_x, center_y, is_buddy=False)
    survivor_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=survivor.rect,
        sprites=[survivor],
        enable_shadows=True,
    )
    survivor_path = out / "survivor.png"
    _save_surface(survivor_surface, survivor_path, scale=output_scale)
    saved.append(survivor_path)

    car = Car(center_x, center_y, appearance="default")
    car._set_facing_bin(0)
    car_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=car.rect,
        sprites=[car],
        enable_shadows=True,
    )
    car_path = out / "car.png"
    _save_surface(car_surface, car_path, scale=output_scale)
    saved.append(car_path)

    patrol_bot = PatrolBot(center_x, center_y)
    patrol_bot_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=patrol_bot.rect,
        sprites=[patrol_bot],
        enable_shadows=True,
    )
    patrol_bot_path = out / "patrol-bot.png"
    _save_surface(patrol_bot_surface, patrol_bot_path, scale=output_scale)
    saved.append(patrol_bot_path)

    fuel = FuelCan(center_x, center_y)
    fuel_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=fuel.rect,
        sprites=[fuel],
    )
    fuel_path = out / "fuel.png"
    _save_surface(fuel_surface, fuel_path, scale=output_scale)
    saved.append(fuel_path)

    empty_fuel_can = EmptyFuelCan(center_x, center_y)
    empty_fuel_can_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=empty_fuel_can.rect,
        sprites=[empty_fuel_can],
    )
    empty_fuel_can_path = out / "empty-fuel-can.png"
    _save_surface(empty_fuel_can_surface, empty_fuel_can_path, scale=output_scale)
    saved.append(empty_fuel_can_path)

    fuel_station = FuelStation(center_x, center_y)
    fuel_station_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=fuel_station.rect,
        sprites=[fuel_station],
    )
    fuel_station_path = out / "fuel-station.png"
    _save_surface(fuel_station_surface, fuel_station_path, scale=output_scale)
    saved.append(fuel_station_path)

    flashlight = Flashlight(center_x, center_y)
    flashlight_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=flashlight.rect,
        sprites=[flashlight],
    )
    flashlight_path = out / "flashlight.png"
    _save_surface(flashlight_surface, flashlight_path, scale=output_scale)
    saved.append(flashlight_path)

    shoes = Shoes(center_x, center_y)
    shoes_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=shoes.rect,
        sprites=[shoes],
    )
    shoes_path = out / "shoes.png"
    _save_surface(shoes_surface, shoes_path, scale=output_scale)
    saved.append(shoes_path)

    beam = SteelBeam(
        center_x - cell_size // 2,
        center_y - cell_size // 2,
        cell_size,
        health=STEEL_BEAM_HEALTH,
        palette=None,
    )
    beam_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=beam.rect,
        sprites=[beam],
    )
    beam_path = out / "steel-beam.png"
    _save_surface(beam_surface, beam_path, scale=output_scale)
    saved.append(beam_path)

    inner_wall = Wall(
        center_x - cell_size // 2,
        center_y - cell_size // 2,
        cell_size,
        cell_size,
        palette_category="inner_wall",
        bevel_depth=INTERNAL_WALL_BEVEL_DEPTH,
        bevel_mask=(True, True, True, True),
        draw_bottom_side=True,
        bottom_side_ratio=0.1,
        side_shade_ratio=0.9,
    )
    inner_wall_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=inner_wall.rect,
        sprites=[inner_wall],
    )
    inner_wall_path = out / "wall-inner.png"
    _save_surface(inner_wall_surface, inner_wall_path, scale=output_scale)
    saved.append(inner_wall_path)

    rubble_wall = RubbleWall(
        center_x - cell_size // 2,
        center_y - cell_size // 2,
        cell_size,
        cell_size,
        palette_category="inner_wall",
    )
    rubble_wall_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=rubble_wall.rect,
        sprites=[rubble_wall],
    )
    rubble_wall_path = out / "wall-rubble.png"
    _save_surface(rubble_wall_surface, rubble_wall_path, scale=output_scale)
    saved.append(rubble_wall_path)

    outer_wall = Wall(
        center_x - cell_size // 2,
        center_y - cell_size // 2,
        cell_size,
        cell_size,
        palette_category="outer_wall",
        bevel_depth=0,
        bevel_mask=(True, True, True, True),
        draw_bottom_side=True,
        bottom_side_ratio=0.1,
        side_shade_ratio=0.9,
    )
    outer_wall_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=outer_wall.rect,
        sprites=[outer_wall],
    )
    outer_wall_path = out / "wall-outer.png"
    _save_surface(outer_wall_surface, outer_wall_path, scale=output_scale)
    saved.append(outer_wall_path)

    reinforced_wall = ReinforcedWall(
        center_x - cell_size // 2,
        center_y - cell_size // 2,
        cell_size,
        cell_size,
        bevel_depth=INTERNAL_WALL_BEVEL_DEPTH,
        bevel_mask=(True, True, True, True),
        draw_bottom_side=True,
        bottom_side_ratio=0.1,
        side_shade_ratio=0.9,
        health=STEEL_BEAM_HEALTH,
    )
    reinforced_wall_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=reinforced_wall.rect,
        sprites=[reinforced_wall],
    )
    legacy_reinforced_wall_path = out / "wall-reinforced-concept.png"
    if legacy_reinforced_wall_path.exists():
        legacy_reinforced_wall_path.unlink()
    reinforced_wall_path = out / "wall-reinforced.png"
    _save_surface(reinforced_wall_surface, reinforced_wall_path, scale=output_scale)
    saved.append(reinforced_wall_path)

    pitfall_rect = pygame.Rect(
        cell_x * cell_size,
        cell_y * cell_size,
        cell_size,
        cell_size,
    )
    pitfall_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=pitfall_rect,
        pitfall_cells={(cell_x, cell_y)},
    )
    pitfall_path = out / "pitfall.png"
    _save_surface(pitfall_surface, pitfall_path, scale=output_scale)
    saved.append(pitfall_path)

    fire_floor_rect = pygame.Rect(
        cell_x * cell_size,
        cell_y * cell_size,
        cell_size,
        cell_size,
    )
    fire_floor_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=fire_floor_rect,
        fire_floor_cells={(cell_x, cell_y)},
    )
    fire_floor_path = out / "fire-floor.png"
    _save_surface(fire_floor_surface, fire_floor_path, scale=output_scale)
    saved.append(fire_floor_path)

    fall_target = (center_x, center_y)
    falling = FallingEntity(
        start_pos=fall_target,
        target_pos=fall_target,
        started_at_ms=0,
        pre_fx_ms=0,
        fall_duration_ms=1000,
        dust_duration_ms=0,
        kind=None,
        mode="pitfall",
    )
    fall_rect = pygame.Rect(0, 0, ZOMBIE_RADIUS * 2, ZOMBIE_RADIUS * 2)
    fall_rect.center = fall_target
    falling_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=fall_rect,
        falling_zombies=[falling],
    )
    falling_path = out / "falling-zombie.png"
    _save_surface(falling_surface, falling_path, scale=output_scale)
    saved.append(falling_path)

    fall_zone_cell_x = max(0, min(cols - 4, cell_x - 2))
    fall_zone_cell_y = max(0, min(rows - 4, cell_y - 2))
    fall_zone_cells = {
        (fall_zone_cell_x + dx, fall_zone_cell_y + dy)
        for dx in range(4)
        for dy in range(4)
    }
    fall_zone_rect = pygame.Rect(
        fall_zone_cell_x * cell_size,
        fall_zone_cell_y * cell_size,
        cell_size * 4,
        cell_size * 4,
    )
    fall_zone_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=fall_zone_rect,
        fall_spawn_cells=fall_zone_cells,
    )
    fall_zone_path = out / "fall-zone.png"
    _save_surface(fall_zone_surface, fall_zone_path, scale=output_scale)
    saved.append(fall_zone_path)

    moving_floor_rect = pygame.Rect(
        cell_x * cell_size,
        cell_y * cell_size,
        cell_size,
        cell_size,
    )
    moving_floor_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=moving_floor_rect,
        moving_floor_cells={(cell_x, cell_y): MovingFloorDirection.RIGHT},
    )
    moving_floor_path = out / "moving-floor.png"
    _save_surface(moving_floor_surface, moving_floor_path, scale=output_scale)
    saved.append(moving_floor_path)

    puddle_cell_x = cell_x
    puddle_cell_y = cell_y
    puddle_rect = pygame.Rect(
        puddle_cell_x * cell_size,
        puddle_cell_y * cell_size,
        cell_size,
        cell_size,
    )
    puddle_padding = max(2, int(round(cell_size * 0.12)))
    puddle_rect = puddle_rect.inflate(puddle_padding * 2, puddle_padding * 2)
    puddle_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=puddle_rect,
        puddle_cells={(puddle_cell_x, puddle_cell_y)},
        ambient_palette_key=None,
    )
    puddle_path = out / "puddle.png"
    _save_surface(puddle_surface, puddle_path, scale=output_scale)
    saved.append(puddle_path)

    spiky_plant = SpikyPlant(center_x, center_y)
    spiky_plant_surface = _render_studio_snapshot(
        cell_size=cell_size,
        target_rect=spiky_plant.rect,
        sprites=[spiky_plant],
        enable_shadows=True,
    )
    spiky_plant_path = out / "spiky_plant.png"
    _save_surface(spiky_plant_surface, spiky_plant_path, scale=output_scale)
    saved.append(spiky_plant_path)

    return saved



from __future__ import annotations

from collections.abc import Iterable
from enum import IntFlag
from typing import TYPE_CHECKING

import pygame

if TYPE_CHECKING:  # pragma: no cover - typing-only imports
    pass


SPATIAL_INDEX_CELL_SIZE = 32


class SpatialKind(IntFlag):
    NONE = 0
    PLAYER = 1 << 0
    CAR = 1 << 1
    ZOMBIE = 1 << 2
    ZOMBIE_DOG = 1 << 3
    TRAPPED_ZOMBIE = 1 << 4
    SURVIVOR = 1 << 5
    PATROL_BOT = 1 << 6
    ALL = PLAYER | CAR | ZOMBIE | ZOMBIE_DOG | TRAPPED_ZOMBIE | SURVIVOR | PATROL_BOT


def _entity_center(entity: pygame.sprite.Sprite) -> tuple[float, float]:
    rect = getattr(entity, "rect", None)
    if rect is not None:
        return float(rect.centerx), float(rect.centery)
    x = getattr(entity, "x", 0.0)
    y = getattr(entity, "y", 0.0)
    return float(x), float(y)


def kind_for_entity(entity: pygame.sprite.Sprite) -> SpatialKind:
    from ..entities import (
        Car,
        PatrolBot,
        Player,
        Survivor,
        Zombie,
        ZombieDog,
        TrappedZombie,
    )

    if isinstance(entity, Player):
        return SpatialKind.PLAYER
    if isinstance(entity, Car):
        return SpatialKind.CAR
    if isinstance(entity, ZombieDog):
        return SpatialKind.ZOMBIE_DOG
    if isinstance(entity, Zombie):
        return SpatialKind.ZOMBIE
    if isinstance(entity, TrappedZombie):
        return SpatialKind.TRAPPED_ZOMBIE
    if isinstance(entity, Survivor):
        return SpatialKind.SURVIVOR
    if isinstance(entity, PatrolBot):
        return SpatialKind.PATROL_BOT
    return SpatialKind.NONE


class SpatialIndex:
    def __init__(self, cell_size: int = SPATIAL_INDEX_CELL_SIZE) -> None:
        self.cell_size = max(1, int(cell_size))
        self._cells: dict[
            tuple[int, int], list[tuple[pygame.sprite.Sprite, SpatialKind]]
        ] = {}

    def clear(self) -> None:
        self._cells.clear()

    def rebuild(self, entities: Iterable[pygame.sprite.Sprite]) -> None:
        self.clear()
        for entity in entities:
            if not getattr(entity, "alive", lambda: True)():
                continue
            kind = kind_for_entity(entity)
            if kind == SpatialKind.NONE:
                continue
            self.insert(entity, kind)

    def insert(self, entity: pygame.sprite.Sprite, kind: SpatialKind) -> None:
        x, y = _entity_center(entity)
        cell = (int(x // self.cell_size), int(y // self.cell_size))
        self._cells.setdefault(cell, []).append((entity, kind))

    def query_radius(
        self,
        center: tuple[float, float],
        radius: float,
        *,
        kinds: SpatialKind = SpatialKind.ALL,
    ) -> list[pygame.sprite.Sprite]:
        if kinds == SpatialKind.NONE:
            return []
        radius = max(0.0, float(radius))
        if radius <= 0:
            return []
        cx, cy = center
        min_x = int((cx - radius) // self.cell_size)
        max_x = int((cx + radius) // self.cell_size)
        min_y = int((cy - radius) // self.cell_size)
        max_y = int((cy + radius) // self.cell_size)
        radius_sq = radius * radius
        results: list[pygame.sprite.Sprite] = []
        seen: set[int] = set()
        for cell_y in range(min_y, max_y + 1):
            for cell_x in range(min_x, max_x + 1):
                bucket = self._cells.get((cell_x, cell_y))
                if not bucket:
                    continue
                for entity, kind in bucket:
                    if kind & kinds == 0:
                        continue
                    ent_id = id(entity)
                    if ent_id in seen:
                        continue
                    ex, ey = _entity_center(entity)
                    dx = ex - cx
                    dy = ey - cy
                    if dx * dx + dy * dy <= radius_sq:
                        results.append(entity)
                        seen.add(ent_id)
        return results

    def query_aabb(
        self,
        rect: pygame.Rect,
        *,
        kinds: SpatialKind = SpatialKind.ALL,
    ) -> list[pygame.sprite.Sprite]:
        if kinds == SpatialKind.NONE:
            return []
        min_x = int(rect.left // self.cell_size)
        max_x = int(rect.right // self.cell_size)
        min_y = int(rect.top // self.cell_size)
        max_y = int(rect.bottom // self.cell_size)
        results: list[pygame.sprite.Sprite] = []
        seen: set[int] = set()
        for cell_y in range(min_y, max_y + 1):
            for cell_x in range(min_x, max_x + 1):
                bucket = self._cells.get((cell_x, cell_y))
                if not bucket:
                    continue
                for entity, kind in bucket:
                    if kind & kinds == 0:
                        continue
                    ent_id = id(entity)
                    if ent_id in seen:
                        continue
                    ent_rect = getattr(entity, "rect", None)
                    if ent_rect is None or rect.colliderect(ent_rect):
                        results.append(entity)
                        seen.add(ent_id)
        return results

    def query_cells(
        self,
        *,
        min_cell_x: int,
        max_cell_x: int,
        min_cell_y: int,
        max_cell_y: int,
        kinds: SpatialKind = SpatialKind.ALL,
    ) -> list[pygame.sprite.Sprite]:
        if kinds == SpatialKind.NONE:
            return []
        if min_cell_x > max_cell_x or min_cell_y > max_cell_y:
            return []
        results: list[pygame.sprite.Sprite] = []
        seen: set[int] = set()
        for cell_y in range(min_cell_y, max_cell_y + 1):
            for cell_x in range(min_cell_x, max_cell_x + 1):
                bucket = self._cells.get((cell_x, cell_y))
                if not bucket:
                    continue
                for entity, kind in bucket:
                    if kind & kinds == 0:
                        continue
                    ent_id = id(entity)
                    if ent_id in seen:
                        continue
                    results.append(entity)
                    seen.add(ent_id)
        return results
